# Copyright (C) 2017-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from datetime import datetime, timezone
from typing import List, Optional

from swh.scheduler.interface import SchedulerInterface
from swh.scheduler.model import (
    ListedOrigin,
    Lister,
    Task,
    TaskArguments,
    TaskPolicy,
    TaskPriority,
)


def utcnow():
    return datetime.now(tz=timezone.utc)


def get_task(task_name):
    """Retrieve task object in our application instance by its fully
    qualified python name.

    Args:
        task_name (str): task's name (e.g
                         swh.loader.git.tasks.LoadDiskGitRepository)

    Returns:
        Instance of task

    """
    from swh.scheduler.celery_backend.config import app

    for module in app.conf.CELERY_IMPORTS:
        __import__(module)
    return app.tasks[task_name]


def create_task(
    type: str,
    policy: TaskPolicy,
    *args,
    next_run: Optional[datetime] = None,
    priority: Optional[TaskPriority] = None,
    retries_left: Optional[int] = None,
    **kwargs,
) -> Task:
    """Create a task with type and policy, scheduled for as soon as
       possible.

    Args:
        type: Type of oneshot task as per swh-scheduler's db
                    table task_type's column (Ex: load-git,
                    check-deposit)
        policy: oneshot or recurring policy
        next_run: optional date and time from which the task can be executed,
            use current time otherwise

    Returns:
        Expected dictionary for the one-shot task scheduling api
        (swh.scheduler.backend.create_tasks)

    """
    task = Task(
        policy=policy,
        type=type,
        next_run=next_run or utcnow(),
        arguments=TaskArguments(
            args=list(args) if args else [],
            kwargs=kwargs if kwargs else {},
        ),
        priority=priority,
        retries_left=retries_left,
    )
    return task


def create_origin_task(origin: ListedOrigin, lister: Lister) -> Task:
    if origin.lister_id != lister.id:
        raise ValueError(
            "origin.lister_id and lister.id differ", origin.lister_id, lister.id
        )
    return Task(
        type=f"load-{origin.visit_type}",
        arguments=TaskArguments(
            args=[],
            kwargs={
                "url": origin.url,
                "lister_name": lister.name,
                "lister_instance_name": lister.instance_name or None,
                **origin.extra_loader_arguments,
            },
        ),
        next_run=utcnow(),
    )


def create_origin_tasks(
    origins: List[ListedOrigin], scheduler: SchedulerInterface
) -> List[Task]:
    """Returns a task dict for each origin, in the same order."""

    lister_ids = {o.lister_id for o in origins}
    listers = {
        lister.id: lister
        for lister in scheduler.get_listers_by_id(list(map(str, lister_ids)))
    }

    missing_lister_ids = lister_ids - set(listers)
    assert not missing_lister_ids, f"Missing listers: {missing_lister_ids}"

    return [create_origin_task(o, listers[o.lister_id]) for o in origins]


def create_oneshot_task(
    type: str, *args, next_run: Optional[datetime] = None, **kwargs
) -> Task:
    """Create a oneshot task scheduled for as soon as possible.

    Args:
        type: Type of oneshot task as per swh-scheduler's db
            table task_type's column (Ex: load-git, check-deposit)
        next_run: optional date and time from which the task can be executed,
            use current time otherwise

    Returns:
        Expected dictionary for the one-shot task scheduling api
        (:func:`swh.scheduler.backend.create_tasks`)

    """
    return create_task(type, "oneshot", *args, next_run=next_run, **kwargs)



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from itertools import islice
import logging
import random
import re
import tempfile
from unittest.mock import patch

from click.testing import CliRunner
import pytest

from swh.core.api.classes import stream_results
from swh.model.model import Origin
from swh.scheduler.cli import cli
from swh.scheduler.utils import create_task, utcnow

CLI_CONFIG = """
scheduler:
    cls: foo
    args: {}
"""


def invoke(scheduler, catch_exceptions, args, config=CLI_CONFIG, **kwargs):
    runner = CliRunner()
    with patch(
        "swh.scheduler.get_scheduler"
    ) as get_scheduler_mock, tempfile.NamedTemporaryFile(
        "a", suffix=".yml"
    ) as config_fd:
        config_fd.write(config)
        config_fd.seek(0)
        get_scheduler_mock.return_value = scheduler
        args = [
            "-C" + config_fd.name,
        ] + args
        result = runner.invoke(cli, args, obj={"log_level": logging.WARNING}, **kwargs)
    if not catch_exceptions and result.exception:
        print(result.output)
        raise result.exception
    return result


def test_schedule_tasks(swh_scheduler):
    csv_data = (
        b'swh-test-ping;[["arg1", "arg2"]];{"key": "value"};'
        + utcnow().isoformat().encode()
        + b"\n"
        + b'swh-test-ping;[["arg3", "arg4"]];{"key": "value"};'
        + utcnow().isoformat().encode()
        + b"\n"
    )
    with tempfile.NamedTemporaryFile(suffix=".csv") as csv_fd:
        csv_fd.write(csv_data)
        csv_fd.seek(0)
        result = invoke(
            swh_scheduler, False, ["task", "schedule", "-d", ";", csv_fd.name]
        )
    expected = r"""
Created 2 tasks

Task 1
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: recurring
  Args:
    \['arg1', 'arg2'\]
  Keyword args:
    key: 'value'

Task 2
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: recurring
  Args:
    \['arg3', 'arg4'\]
  Keyword args:
    key: 'value'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_schedule_tasks_columns(swh_scheduler):
    with tempfile.NamedTemporaryFile(suffix=".csv") as csv_fd:
        csv_fd.write(b'swh-test-ping;oneshot;["arg1", "arg2"];{"key": "value"}\n')
        csv_fd.seek(0)
        result = invoke(
            swh_scheduler,
            False,
            [
                "task",
                "schedule",
                "-c",
                "type",
                "-c",
                "policy",
                "-c",
                "args",
                "-c",
                "kwargs",
                "-d",
                ";",
                csv_fd.name,
            ],
        )
    expected = r"""
Created 1 tasks

Task 1
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
    'arg1'
    'arg2'
  Keyword args:
    key: 'value'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_schedule_task(swh_scheduler):
    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "add",
            "swh-test-ping",
            "arg1",
            "arg2",
            "key=value",
        ],
    )
    expected = r"""
Created 1 tasks

Task 1
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: recurring
  Args:
    'arg1'
    'arg2'
  Keyword args:
    key: 'value'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_schedule_unknown_task_type(swh_scheduler):
    """When scheduling unknown task type, the cli should raise."""
    result = invoke(
        swh_scheduler,
        True,
        [
            "task",
            "add",
            "unknown-task-type-should-raise",
        ],
    )

    assert "Unknown" in result.output
    assert result.exit_code != 0


def test_list_pending_tasks_none(swh_scheduler):
    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list-pending",
            "swh-test-ping",
        ],
    )

    expected = r"""
Found 0 swh-test-ping tasks

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_pending_tasks(swh_scheduler):
    task1 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow(),
        key="value1",
    )
    task2 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow() + datetime.timedelta(days=1),
        key="value2",
    )
    swh_scheduler.create_tasks([task1, task2])

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list-pending",
            "swh-test-ping",
        ],
    )

    expected = r"""
Found 1 swh-test-ping tasks

Task 1
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value1'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    swh_scheduler.grab_ready_tasks("swh-test-ping")

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list-pending",
            "swh-test-ping",
        ],
    )

    expected = r"""
Found 0 swh-test-ping tasks

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_pending_tasks_filter(swh_scheduler):
    task = create_task("swh-test-multiping", "oneshot", key="value")
    swh_scheduler.create_tasks([task])

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list-pending",
            "swh-test-ping",
        ],
    )

    expected = r"""
Found 0 swh-test-ping tasks

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_pending_tasks_filter_2(swh_scheduler):
    swh_scheduler.create_tasks(
        [
            create_task("swh-test-multiping", "oneshot", key="value"),
            create_task("swh-test-ping", "oneshot", key="value2"),
        ]
    )

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list-pending",
            "swh-test-ping",
        ],
    )

    expected = r"""
Found 1 swh-test-ping tasks

Task 2
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value2'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


# Fails because "task list-pending --limit 3" only returns 2 tasks, because
# of how compute_nb_tasks_from works.
@pytest.mark.xfail
def test_list_pending_tasks_limit(swh_scheduler):
    swh_scheduler.create_tasks(
        [create_task("swh-test-ping", "oneshot", key="value%d" % i) for i in range(10)]
    )

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list-pending",
            "swh-test-ping",
            "--limit",
            "3",
        ],
    )

    expected = r"""
Found 2 swh-test-ping tasks

Task 1
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value0'

Task 2
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value1'

Task 3
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value2'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_pending_tasks_before(swh_scheduler):
    task1 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow() + datetime.timedelta(days=3),
        key="value",
    )
    task2 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow() + datetime.timedelta(days=1),
        key="value2",
    )

    swh_scheduler.create_tasks([task1, task2])

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list-pending",
            "swh-test-ping",
            "--before",
            (datetime.date.today() + datetime.timedelta(days=2)).isoformat(),
        ],
    )

    expected = r"""
Found 1 swh-test-ping tasks

Task 2
  Next run: tomorrow \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value2'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks(swh_scheduler):
    task1 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow() + datetime.timedelta(days=3, hours=2),
        key="value1",
    )
    task2 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow(),
        key="value2",
    )
    swh_scheduler.create_tasks([task1, task2])

    swh_scheduler.grab_ready_tasks("swh-test-ping")

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list",
        ],
    )

    expected = r"""
Found 2 tasks

Task 1
  Next run: .+ \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value1'

Task 2
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_id(swh_scheduler):
    task1 = create_task("swh-test-ping", "oneshot", key="value1")
    task2 = create_task("swh-test-ping", "oneshot", key="value2")
    task3 = create_task("swh-test-ping", "oneshot", key="value3")
    swh_scheduler.create_tasks([task1, task2, task3])

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list",
            "--task-id",
            "2",
        ],
    )

    expected = r"""
Found 1 tasks

Task 2
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_id_2(swh_scheduler):
    task1 = create_task("swh-test-ping", "oneshot", key="value1")
    task2 = create_task("swh-test-ping", "oneshot", key="value2")
    task3 = create_task("swh-test-ping", "oneshot", key="value3")
    swh_scheduler.create_tasks([task1, task2, task3])

    result = invoke(
        swh_scheduler, False, ["task", "list", "--task-id", "2", "--task-id", "3"]
    )

    expected = r"""
Found 2 tasks

Task 2
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

Task 3
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value3'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_type(swh_scheduler):
    task1 = create_task("swh-test-ping", "oneshot", key="value1")
    task2 = create_task("swh-test-multiping", "oneshot", key="value2")
    task3 = create_task("swh-test-ping", "oneshot", key="value3")
    swh_scheduler.create_tasks([task1, task2, task3])

    result = invoke(
        swh_scheduler, False, ["task", "list", "--task-type", "swh-test-ping"]
    )

    expected = r"""
Found 2 tasks

Task 1
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value1'

Task 3
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value3'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_limit(swh_scheduler):
    task1 = create_task("swh-test-ping", "oneshot", key="value1")
    task2 = create_task("swh-test-ping", "oneshot", key="value2")
    task3 = create_task("swh-test-ping", "oneshot", key="value3")
    swh_scheduler.create_tasks([task1, task2, task3])

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list",
            "--limit",
            "2",
        ],
    )

    expected = r"""
Found 2 tasks

Task 1
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value1'

Task 2
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_before(swh_scheduler):
    task1 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow() + datetime.timedelta(days=3, hours=2),
        key="value1",
    )
    task2 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow(),
        key="value2",
    )
    swh_scheduler.create_tasks([task1, task2])

    swh_scheduler.grab_ready_tasks("swh-test-ping")

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list",
            "--before",
            (datetime.date.today() + datetime.timedelta(days=2)).isoformat(),
        ],
    )

    expected = r"""
Found 1 tasks

Task 2
  Next run: today \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_after(swh_scheduler):
    task1 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow() + datetime.timedelta(days=3, hours=2),
        key="value1",
    )
    task2 = create_task(
        "swh-test-ping",
        "oneshot",
        next_run=utcnow(),
        key="value2",
    )
    swh_scheduler.create_tasks([task1, task2])

    swh_scheduler.grab_ready_tasks("swh-test-ping")

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "list",
            "--after",
            (datetime.date.today() + datetime.timedelta(days=2)).isoformat(),
        ],
    )

    expected = r"""
Found 1 tasks

Task 1
  Next run: .+ \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value1'

""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def _fill_storage_with_origins(storage, nb_origins):
    origins = [Origin(url=f"http://example.com/{i}") for i in range(nb_origins)]
    storage.origin_add(origins)
    return origins


@patch("swh.scheduler.cli.utils.TASK_BATCH_SIZE", 3)
def test_task_schedule_origins_dry_run(swh_scheduler, storage):
    """Tests the scheduling when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    _fill_storage_with_origins(storage, 90)

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "schedule_origins",
            "--dry-run",
            "swh-test-ping",
        ],
    )

    # Check the output
    expected = r"""
Scheduled 3 tasks \(30 origins\).
Scheduled 6 tasks \(60 origins\).
Scheduled 9 tasks \(90 origins\).
Done.
""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), repr(result.output)

    # Check scheduled tasks
    tasks = swh_scheduler.search_tasks()
    assert len(tasks) == 0


def _assert_origin_tasks_contraints(tasks, max_tasks, max_task_size, expected_origins):
    # check there are not too many tasks
    assert len(tasks) <= max_tasks

    # check tasks are not too large
    assert all(len(task.arguments.args[0]) <= max_task_size for task in tasks)

    # check the tasks are exhaustive
    assert sum([len(task.arguments.args[0]) for task in tasks]) == len(expected_origins)
    assert set.union(*(set(task.arguments.args[0]) for task in tasks)) == {
        origin.url for origin in expected_origins
    }


@patch("swh.scheduler.cli.utils.TASK_BATCH_SIZE", 3)
def test_task_schedule_origins(swh_scheduler, storage):
    """Tests the scheduling when neither origin_batch_size or
    task_batch_size is a divisor of nb_origins."""
    origins = _fill_storage_with_origins(storage, 70)

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "schedule_origins",
            "swh-test-ping",
            "--batch-size",
            "20",
        ],
    )

    # Check the output
    expected = r"""
Scheduled 3 tasks \(60 origins\).
Scheduled 4 tasks \(70 origins\).
Done.
""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), repr(result.output)

    # Check tasks
    tasks = swh_scheduler.search_tasks()
    _assert_origin_tasks_contraints(tasks, 4, 20, origins)
    assert all(task.arguments.kwargs == {} for task in tasks)


def test_task_schedule_origins_kwargs(swh_scheduler, storage):
    """Tests support of extra keyword-arguments."""
    origins = _fill_storage_with_origins(storage, 30)

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "schedule_origins",
            "swh-test-ping",
            "--batch-size",
            "20",
            'key1="value1"',
            'key2="value2"',
        ],
    )

    # Check the output
    expected = r"""
Scheduled 2 tasks \(30 origins\).
Done.
""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), repr(result.output)

    # Check tasks
    tasks = swh_scheduler.search_tasks()
    _assert_origin_tasks_contraints(tasks, 2, 20, origins)
    assert all(
        task.arguments.kwargs == {"key1": "value1", "key2": "value2"} for task in tasks
    )


def test_task_schedule_origins_with_limit(swh_scheduler, storage):
    """Tests support of extra keyword-arguments."""
    _fill_storage_with_origins(storage, 50)
    limit = 20
    expected_origins = list(islice(stream_results(storage.origin_list), limit))
    nb_origins = len(expected_origins)

    assert nb_origins == limit
    max_task_size = 5
    nb_tasks, remainder = divmod(nb_origins, max_task_size)
    assert remainder == 0  # made the numbers go round

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "schedule_origins",
            "swh-test-ping",
            "--batch-size",
            max_task_size,
            "--limit",
            limit,
        ],
    )

    # Check the output
    expected = rf"""
Scheduled {nb_tasks} tasks \({nb_origins} origins\).
Done.
""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), repr(result.output)

    tasks = swh_scheduler.search_tasks()
    _assert_origin_tasks_contraints(tasks, max_task_size, nb_origins, expected_origins)


def test_task_schedule_origins_with_page_token(swh_scheduler, storage):
    """Tests support of extra keyword-arguments."""
    nb_total_origins = 50
    origins = _fill_storage_with_origins(storage, nb_total_origins)

    # prepare page_token and origins result expectancy
    page_result = storage.origin_list(limit=10)
    assert len(page_result.results) == 10
    page_token = page_result.next_page_token
    assert page_token is not None

    # remove the first 10 origins listed as we won't see those in tasks
    expected_origins = [o for o in origins if o not in page_result.results]
    nb_origins = len(expected_origins)
    assert nb_origins == nb_total_origins - len(page_result.results)

    max_task_size = 10
    nb_tasks, remainder = divmod(nb_origins, max_task_size)
    assert remainder == 0

    result = invoke(
        swh_scheduler,
        False,
        [
            "task",
            "schedule_origins",
            "swh-test-ping",
            "--batch-size",
            max_task_size,
            "--page-token",
            page_token,
        ],
    )

    # Check the output
    expected = rf"""
Scheduled {nb_tasks} tasks \({nb_origins} origins\).
Done.
""".lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), repr(result.output)

    # Check tasks
    tasks = swh_scheduler.search_tasks()
    _assert_origin_tasks_contraints(tasks, max_task_size, nb_origins, expected_origins)


def test_cli_task_runner_unknown_task_types(swh_scheduler, storage):
    """When passing at least one unknown task type, the runner should fail."""

    task_types = swh_scheduler.get_task_types()
    task_type_names = [t.type for t in task_types]
    known_task_type = random.choice(task_type_names)
    unknown_task_type = "unknown-task-type"
    assert unknown_task_type not in task_type_names

    result = invoke(
        swh_scheduler,
        True,
        [
            "start-runner",
            "--task-type",
            known_task_type,
            "--task-type",
            unknown_task_type,
        ],
    )
    assert "Unknown" in result.output
    assert result.exit_code != 0


@pytest.mark.parametrize("flag_priority", ["--with-priority", "--without-priority"])
def test_cli_task_runner_with_known_tasks(
    swh_scheduler, storage, caplog, flag_priority
):
    """Trigger runner with known tasks runs smoothly."""

    task_types = swh_scheduler.get_task_types()
    task_type_names = [t.type for t in task_types]
    task_type_name = random.choice(task_type_names)
    task_type_name2 = random.choice(task_type_names)

    # The runner will just iterate over the following known tasks and do noop. We are
    # just checking the runner does not explode here.
    result = invoke(
        swh_scheduler,
        False,
        [
            "start-runner",
            flag_priority,
            "--task-type",
            task_type_name,
            "--task-type",
            task_type_name2,
        ],
    )

    assert result.exit_code == 0, result.output


def test_cli_task_runner_no_task(swh_scheduler, storage):
    """Trigger runner with no parameter should run as before."""

    # The runner will just iterate over the existing tasks from the scheduler and do
    # noop. We are just checking the runner does not explode here.
    result = invoke(
        swh_scheduler,
        True,
        [
            "start-runner",
        ],
    )

    assert result.exit_code == 0, result.output



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import traceback

from click.testing import CliRunner
import pytest
import yaml

from swh.scheduler import get_scheduler
from swh.scheduler.cli import cli

FAKE_MODULE_ENTRY_POINTS = {
    "lister.foo=swh.scheduler.tests.fixtures.lister.foo:register",
    "loader.bar=swh.scheduler.tests.fixtures.loader.bar:register",
}


@pytest.fixture
def cli_runner():
    return CliRunner()


@pytest.fixture(autouse=True)
def mock_plugin_worker_descriptions(mocker):
    """Register fake lister and loader for testing celery tasks registration."""
    from pkg_resources import Distribution, EntryPoint

    d = Distribution()
    entry_points = [
        EntryPoint.parse(entry, dist=d) for entry in FAKE_MODULE_ENTRY_POINTS
    ]
    mocker.patch(
        "swh.scheduler.cli.task_type._plugin_worker_descriptions",
    ).return_value = {entry_point.name: entry_point for entry_point in entry_points}


@pytest.fixture
def local_sched_config(swh_scheduler_config):
    """Expose the local scheduler configuration"""
    return {"scheduler": {"cls": "local", **swh_scheduler_config}}


@pytest.fixture
def local_sched_configfile(local_sched_config, tmp_path):
    """Write in temporary location the local scheduler configuration"""
    configfile = tmp_path / "config.yml"
    configfile.write_text(yaml.dump(local_sched_config))
    return configfile.as_posix()


def test_register_unknown_task_type(
    cli_runner,
    local_sched_configfile,
):
    """Trying to register an unknown task type should return an error."""

    command = [
        "--config-file",
        local_sched_configfile,
        "task-type",
        "register",
        "-p",
        "lister.baz",
    ]

    result = cli_runner.invoke(cli, command)

    assert result.exit_code != 0
    assert (
        "That provided plugin is unknown: lister.baz.\n"
        "Available ones are: lister.foo, loader.bar."
    ) in result.output


def test_register_task_types_all(
    cli_runner,
    local_sched_config,
    local_sched_configfile,
):
    """Registering all task types"""

    for command in [
        ["--config-file", local_sched_configfile, "task-type", "register"],
        ["--config-file", local_sched_configfile, "task-type", "register", "-p", "all"],
        [
            "--config-file",
            local_sched_configfile,
            "task-type",
            "register",
            "-p",
            "lister.foo",
            "-p",
            "loader.bar",
        ],
    ]:
        result = cli_runner.invoke(cli, command)

        assert result.exit_code == 0, traceback.print_exception(*result.exc_info)

        scheduler = get_scheduler(**local_sched_config["scheduler"])

        all_task_type_names = ["list-foo-full", "load-bar"]
        for task_type_name in all_task_type_names:
            task_type = scheduler.get_task_type(task_type_name)
            assert task_type
            assert task_type.type == task_type_name
            assert task_type.backoff_factor == 1.0


def test_register_task_types_filter(
    cli_runner,
    local_sched_config,
    local_sched_configfile,
):
    """Filtering on one worker should only register its associated task type"""
    result = cli_runner.invoke(
        cli,
        [
            "--config-file",
            local_sched_configfile,
            "task-type",
            "register",
            "--plugins",
            "lister.foo",
        ],
    )

    assert result.exit_code == 0, traceback.print_exception(*result.exc_info)

    scheduler = get_scheduler(**local_sched_config["scheduler"])

    task_type_name = "list-foo-full"
    task_type = scheduler.get_task_type(task_type_name)
    assert task_type
    assert task_type.type == task_type_name
    assert task_type.backoff_factor == 1.0


@pytest.mark.parametrize("cli_command", ["list", "register", "add"])
def test_cli_task_type_raise(cli_runner, cli_command):
    """Without a proper configuration, the cli raises"""
    result = cli_runner.invoke(cli, ["task-type", cli_command])
    assert "Scheduler class" in result.output
    assert result.exit_code != 0



# Copyright (C) 2017-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta, timezone
import uuid

import pytest

from swh.scheduler.model import ListedOrigin, Lister, Task, TaskArguments
from swh.scheduler.utils import (
    create_oneshot_task,
    create_origin_task,
    create_origin_tasks,
    create_task,
    utcnow,
)

from .common import LISTERS


@pytest.fixture
def mock_datetime(mocker):
    now = datetime.now(tz=timezone.utc)
    mock_datetime = mocker.patch("swh.scheduler.utils.datetime")
    mock_datetime.now.return_value = now
    return mock_datetime


def test_create_oneshot_task_simple(mock_datetime):
    actual_task = create_oneshot_task("some-task-type")

    assert actual_task == Task(
        policy="oneshot",
        type="some-task-type",
        next_run=mock_datetime.now.return_value,
        arguments=TaskArguments(
            args=[],
            kwargs={},
        ),
    )
    mock_datetime.now.assert_called_once_with(tz=timezone.utc)


def test_create_oneshot_task_with_next_run():
    next_run = utcnow() + timedelta(hours=1)
    actual_task = create_oneshot_task("some-task-type", next_run=next_run)

    assert actual_task == Task(
        policy="oneshot",
        type="some-task-type",
        next_run=next_run,
        arguments=TaskArguments(
            args=[],
            kwargs={},
        ),
    )


def test_create_oneshot_task_other_call(mock_datetime):
    actual_task = create_oneshot_task(
        "some-task-type", "arg0", "arg1", priority="high", other_stuff="normal"
    )

    assert actual_task == Task(
        policy="oneshot",
        type="some-task-type",
        next_run=mock_datetime.now.return_value,
        arguments=TaskArguments(
            args=["arg0", "arg1"],
            kwargs={"other_stuff": "normal"},
        ),
        priority="high",
    )
    mock_datetime.now.assert_called_once_with(tz=timezone.utc)


def test_create_task(mock_datetime):
    actual_task = create_task(
        "task-type",
        "recurring",
        "arg0",
        "arg1",
        priority="low",
        other_stuff="normal",
        retries_left=3,
    )

    assert actual_task == Task(
        policy="recurring",
        type="task-type",
        next_run=mock_datetime.now.return_value,
        arguments=TaskArguments(
            args=["arg0", "arg1"],
            kwargs={"other_stuff": "normal"},
        ),
        priority="low",
        retries_left=3,
    )
    mock_datetime.now.assert_called_once_with(tz=timezone.utc)


def test_create_origin_task(mock_datetime):
    lister = Lister(**LISTERS[1], id=uuid.uuid4())
    origin = ListedOrigin(
        lister_id=lister.id,
        url="http://example.com/",
        visit_type="git",
    )

    assert create_origin_task(origin, lister) == Task(
        type="load-git",
        next_run=mock_datetime.now.return_value,
        arguments=TaskArguments(
            args=[],
            kwargs={
                "url": "http://example.com/",
                "lister_name": LISTERS[1]["name"],
                "lister_instance_name": LISTERS[1]["instance_name"],
            },
        ),
    )

    loader_args = {"foo": "bar", "baz": {"foo": "bar"}}

    origin_w_args = ListedOrigin(
        lister_id=lister.id,
        url="http://example.com/svn/",
        visit_type="svn",
        extra_loader_arguments=loader_args,
    )

    assert create_origin_task(origin_w_args, lister) == Task(
        type="load-svn",
        next_run=mock_datetime.now.return_value,
        arguments=TaskArguments(
            args=[],
            kwargs={
                "url": "http://example.com/svn/",
                "lister_name": LISTERS[1]["name"],
                "lister_instance_name": LISTERS[1]["instance_name"],
                **loader_args,
            },
        ),
    )


def test_create_origin_tasks(swh_scheduler, mock_datetime):
    listers = []
    for lister_args in LISTERS:
        listers.append(swh_scheduler.get_or_create_lister(**lister_args))

    origin1 = ListedOrigin(
        lister_id=listers[0].id,
        url="http://example.com/1",
        visit_type="git",
    )
    origin2 = ListedOrigin(
        lister_id=listers[0].id,
        url="http://example.com/2",
        visit_type="git",
    )
    origin3 = ListedOrigin(
        lister_id=listers[1].id,
        url="http://example.com/3",
        visit_type="git",
    )

    origins = [origin1, origin2, origin3]

    tasks = create_origin_tasks(origins, swh_scheduler)
    assert tasks == [
        Task(
            type="load-git",
            next_run=mock_datetime.now.return_value,
            arguments=TaskArguments(
                args=[],
                kwargs={
                    "url": "http://example.com/1",
                    "lister_name": LISTERS[0]["name"],
                    "lister_instance_name": LISTERS[0]["instance_name"],
                },
            ),
        ),
        Task(
            type="load-git",
            next_run=mock_datetime.now.return_value,
            arguments=TaskArguments(
                args=[],
                kwargs={
                    "url": "http://example.com/2",
                    "lister_name": LISTERS[0]["name"],
                    "lister_instance_name": LISTERS[0]["instance_name"],
                },
            ),
        ),
        Task(
            type="load-git",
            next_run=mock_datetime.now.return_value,
            arguments=TaskArguments(
                args=[],
                kwargs={
                    "url": "http://example.com/3",
                    "lister_name": LISTERS[1]["name"],
                    "lister_instance_name": LISTERS[1]["instance_name"],
                },
            ),
        ),
    ]



# Copyright (C) 2019-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control

from __future__ import annotations

from datetime import datetime, timedelta
import re
from typing import TYPE_CHECKING, Iterable

import click

from swh.scheduler.model import Task

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Tuple

    from swh.scheduler.interface import SchedulerInterface
    from swh.scheduler.model import TaskPolicy, TaskPriority, TaskRun, TaskType


TASK_BATCH_SIZE = 1000  # Number of tasks per query to the scheduler


TIME_INTERVAL_REGEXP = re.compile(
    r"""
    # optional group for days
    (?:
        (?P<days>\d+\.?|\d*\.\d+) # floating point number, e.g. "1", "2.", "3.4" or ".5".
        \x20*
        (?:day|days|d)
    )?
    \x20? # optional space
    # optional group for hours
    (?:
        (?P<hours>\d+\.?|\d*\.\d+)
        \x20*
        (?:h|hr|hrs|hour|hours)
    )?
    \x20? # optional space
    # optional group for minutes
    (?:
        (?P<minutes>\d+\.?|\d*\.\d+)
        \x20*
        (?:m|min|mins|minute|minutes)
    )?
    \x20? # optional space
    # optional group for seconds
    (?:
        (?P<seconds>\d+\.?|\d*\.\d+)
        \x20*
        (?:s|sec|second|seconds)
    )?
    """,
    re.VERBOSE,
)


def parse_time_interval(time_str: str) -> timedelta:
    """Parse a basic time interval e.g. '1 day' or '2 hours' into a timedelta object.

    Args:
        time_str: A string representing a basic time interval in days or hours.

    Returns:
        An equivalent representation of the string as a datetime.timedelta object.

    Raises:
        ValueError if the time interval could not be parsed.

    """
    parts = TIME_INTERVAL_REGEXP.fullmatch(time_str)
    if not parts:
        raise ValueError(f"{time_str!r} could not be parsed as a time interval")
    time_params = {
        name: float(param) for name, param in parts.groupdict().items() if param
    }
    if not time_params:
        # The regexp lets a bare space go through
        raise ValueError(f"{time_str!r} could not be parsed as a time interval")
    return timedelta(**time_params)


def schedule_origin_batches(scheduler, task_type, origins, origin_batch_size, kwargs):
    from itertools import islice

    from swh.scheduler.utils import create_task

    nb_origins = 0
    nb_tasks = 0

    while True:
        task_batch = []
        for _ in range(TASK_BATCH_SIZE):
            # Group origins
            origin_batch = []
            for origin in islice(origins, origin_batch_size):
                origin_batch.append(origin)
            nb_origins += len(origin_batch)
            if not origin_batch:
                break

            # Create a task for these origins
            args = [origin_batch]
            task_dict = create_task(task_type, "oneshot", *args, **kwargs)
            task_batch.append(task_dict)

        # Schedule a batch of tasks
        if not task_batch:
            break
        nb_tasks += len(task_batch)
        if scheduler:
            scheduler.create_tasks(task_batch)
        click.echo("Scheduled %d tasks (%d origins)." % (nb_tasks, nb_origins))

    # Print final status.
    if nb_tasks:
        click.echo("Done.")
    else:
        click.echo("Nothing to do (no origin metadata matched the criteria).")


def parse_argument(option):
    import yaml

    if option == "":
        # yaml.safe_load("") returns None
        return ""

    try:
        return yaml.safe_load(option)
    except Exception:
        raise click.ClickException("Invalid argument: {}".format(option))


def parse_options(options: List[str]) -> Tuple[List[str], Dict]:
    """Parses options from a CLI as YAML and turns it into Python
    args and kwargs.

    >>> parse_options([])
    ([], {})
    >>> parse_options(['foo', 'bar'])
    (['foo', 'bar'], {})
    >>> parse_options(['[foo, bar]'])
    ([['foo', 'bar']], {})
    >>> parse_options(['"foo"', '"bar"'])
    (['foo', 'bar'], {})
    >>> parse_options(['foo="bar"'])
    ([], {'foo': 'bar'})
    >>> parse_options(['"foo"', 'bar="baz"'])
    (['foo'], {'bar': 'baz'})
    >>> parse_options(['42', 'bar=False'])
    ([42], {'bar': False})
    >>> parse_options(['42', 'bar=false'])
    ([42], {'bar': False})
    >>> parse_options(['foo', ''])
    (['foo', ''], {})
    >>> parse_options(['foo', 'bar='])
    (['foo'], {'bar': ''})
    >>> parse_options(['foo', 'null'])
    (['foo', None], {})
    >>> parse_options(['foo', 'bar=null'])
    (['foo'], {'bar': None})
    >>> parse_options(['42', '"foo'])
    Traceback (most recent call last):
      ...
    click.exceptions.ClickException: Invalid argument: "foo
    """
    kw_pairs = [x.split("=", 1) for x in options if "=" in x]
    args = [parse_argument(x) for x in options if "=" not in x]
    kw = {k: parse_argument(v) for (k, v) in kw_pairs}
    return (args, kw)


def get_task_type(scheduler: SchedulerInterface, visit_type: str) -> Optional[TaskType]:
    "Given a visit type, return its associated task type."
    return scheduler.get_task_type(f"load-{visit_type}")


def send_to_celery(
    scheduler: SchedulerInterface,
    visit_type_to_queue: Dict[str, str],
    enabled: bool = True,
    lister_name: Optional[str] = None,
    lister_instance_name: Optional[str] = None,
    policy: str = "oldest_scheduled_first",
    tablesample: Optional[float] = None,
    absolute_cooldown: Optional[timedelta] = None,
    scheduled_cooldown: Optional[timedelta] = None,
    failed_cooldown: Optional[timedelta] = None,
    not_found_cooldown: Optional[timedelta] = None,
):
    """Utility function to read tasks from the scheduler and send those directly to
    celery.

    Args:
        visit_type_to_queue: Optional mapping of visit/loader type (e.g git, svn, ...)
          to queue to send task to.
        enabled: Determine whether we want to list enabled or disabled origins. As
          default, we want reasonably enabled origins. For some edge case, we might
          want the others.
        lister_name: Determine the list of origins listed from the lister with name
        lister_instance_name: Determine the list of origins listed from the lister
          with instance name
        policy: the scheduling policy used to select which visits to schedule
        tablesample: the percentage of the table on which we run the query
          (None: no sampling)
        absolute_cooldown: the minimal interval between two visits of the same origin
        scheduled_cooldown: the minimal interval before which we can schedule
          the same origin again if it's not been visited
        failed_cooldown: the minimal interval before which we can reschedule a
          failed origin
        not_found_cooldown: the minimal interval before which we can reschedule a
          not_found origin

    """

    from kombu.utils.uuid import uuid

    from swh.scheduler.celery_backend.config import app, get_available_slots

    from ..utils import create_origin_tasks

    for visit_type_name, queue_name in visit_type_to_queue.items():
        task_type = get_task_type(scheduler, visit_type_name)
        assert task_type is not None
        task_name = task_type.backend_name
        num_tasks = get_available_slots(app, queue_name, task_type.max_queue_length)

        click.echo(f"{num_tasks} slots available in celery queue")

        origins = scheduler.grab_next_visits(
            visit_type_name,
            num_tasks,
            policy=policy,
            tablesample=tablesample,
            enabled=enabled,
            lister_name=lister_name,
            lister_instance_name=lister_instance_name,
            absolute_cooldown=absolute_cooldown,
            scheduled_cooldown=scheduled_cooldown,
            failed_cooldown=failed_cooldown,
            not_found_cooldown=not_found_cooldown,
        )

        click.echo(f"{len(origins)} visits to send to celery")
        for task in create_origin_tasks(origins, scheduler):
            app.send_task(
                task_name,
                task_id=uuid(),
                args=task.arguments.args,
                kwargs=task.arguments.kwargs,
                queue=queue_name,
            )


def pretty_print_list(list: List[Any], indent: int = 0):
    """Pretty-print a list"""
    return "".join("%s%r\n" % (" " * indent, item) for item in list)


def pretty_print_dict(dict: Dict[str, Any], indent: int = 0):
    """Pretty-print a list"""
    return "".join(
        "%s%s: %r\n" % (" " * indent, click.style(key, bold=True), value)
        for key, value in sorted(dict.items())
    )


def format_dict(d):
    """Recursively format date objects in the dict passed as argument"""
    import datetime

    ret = {}
    for k, v in d.items():
        if isinstance(v, (datetime.date, datetime.datetime)):
            v = v.isoformat()
        elif isinstance(v, dict):
            v = format_dict(v)
        ret[k] = v
    return ret


def pretty_print_run(run: TaskRun, indent: int = 4):
    fmt = (
        "{indent}{backend_id} [{status}]\n"
        "{indent}  scheduled: {scheduled} [{started}:{ended}]"
    )
    return fmt.format(indent=" " * indent, **format_dict(run.to_dict()))


def pretty_print_task(task: Task, full: bool = False):
    """Pretty-print a task

    If ``full`` is :const:`True`, also print the status and priority fields.

    >>> import datetime
    >>> from swh.scheduler.model import Task, TaskArguments
    >>> task = Task(
    ...     id=1234,
    ...     arguments=TaskArguments(
    ...         args=["foo", "bar", True],
    ...         kwargs={"key": "value", "key2": 42},
    ...     ),
    ...     current_interval=datetime.timedelta(hours=1),
    ...     next_run=datetime.datetime(2019, 2, 21, 13, 52, 35, 407818),
    ...     policy="oneshot",
    ...     priority=None,
    ...     status="next_run_not_scheduled",
    ...     type="test_task",
    ... )
    >>> print(click.unstyle(pretty_print_task(task)))
    Task 1234
      Next run: ... (2019-02-21T13:52:35.407818)
      Interval: 1:00:00
      Type: test_task
      Policy: oneshot
      Args:
        'foo'
        'bar'
        True
      Keyword args:
        key: 'value'
        key2: 42
    <BLANKLINE>
    >>> print(click.unstyle(pretty_print_task(task, full=True)))
    Task 1234
      Next run: ... (2019-02-21T13:52:35.407818)
      Interval: 1:00:00
      Type: test_task
      Policy: oneshot
      Status: next_run_not_scheduled
      Priority:\x20
      Args:
        'foo'
        'bar'
        True
      Keyword args:
        key: 'value'
        key2: 42
    <BLANKLINE>
    """
    import humanize

    next_run = task.next_run
    lines = [
        "%s %s\n" % (click.style("Task", bold=True), task.id),
        click.style("  Next run: ", bold=True),
        "%s (%s)" % (humanize.naturaldate(next_run), next_run.isoformat()),
        "\n",
        click.style("  Interval: ", bold=True),
        str(task.current_interval),
        "\n",
        click.style("  Type: ", bold=True),
        task.type or "",
        "\n",
        click.style("  Policy: ", bold=True),
        task.policy or "",
        "\n",
    ]
    if full:
        lines += [
            click.style("  Status: ", bold=True),
            task.status or "",
            "\n",
            click.style("  Priority: ", bold=True),
            task.priority or "",
            "\n",
        ]
    lines += [
        click.style("  Args:\n", bold=True),
        pretty_print_list(task.arguments.args, indent=4),
        click.style("  Keyword args:\n", bold=True),
        pretty_print_dict(task.arguments.kwargs, indent=4),
    ]

    return "".join(lines)


def task_add(
    scheduler: SchedulerInterface,
    task_type_name: str,
    args: List[str],
    kw: Dict,
    policy: TaskPolicy,
    priority: Optional[TaskPriority] = None,
    next_run: Optional[datetime] = None,
):
    """Add a task task_type_name in the scheduler."""
    from swh.scheduler.model import TaskArguments
    from swh.scheduler.utils import utcnow

    task = Task(
        type=task_type_name,
        policy=policy,
        priority=priority,
        arguments=TaskArguments(
            args=args,
            kwargs=kw,
        ),
        next_run=next_run or utcnow(),
    )
    created = scheduler.create_tasks([task])

    output = [f"Created {len(created)} tasks\n"]
    for task in created:
        output.append(pretty_print_task(task))

    click.echo("\n".join(output))


def lister_task_type(lister_name: str, lister_type: Optional[str] = None) -> str:
    """Compute expected scheduler task type from the lister name and its optional
    listing type (full, incremental).

    """
    prefix = f"list-{lister_name}"
    return f"{prefix}-{lister_type}" if lister_type else prefix


def check_listed_origins(
    scheduler: SchedulerInterface,
    lister_name: str,
    instance_name: str,
    limit: int = 100000,
):
    listed_origins_lister = scheduler.get_lister(
        name=lister_name, instance_name=instance_name
    )

    if listed_origins_lister is None:
        exit(
            f"Forge {instance_name} ({lister_name}) isn't registered \
in the scheduler database."
        )
    else:
        lister_id = listed_origins_lister.id

    listed_origins = scheduler.get_listed_origins(
        lister_id=lister_id,
        enabled=None,
        limit=limit,
    ).results

    if len(listed_origins) == 0:
        exit(
            f"Forge {instance_name} ({lister_name}) has {len(listed_origins)} \
listed origin in the scheduler database."
        )

    return listed_origins


def count_ingested_origins(
    scheduler: SchedulerInterface,
    ids: Iterable[Tuple[str, str]],
    instance_name: str,
    with_listing: Optional[bool] = False,
) -> Tuple[Dict[str, int], List]:
    """Count number of ingested origins grouped by status."""

    ingested_origins = scheduler.origin_visit_stats_get(ids=ids)
    status_counters = {
        "failed": 0,
        "None": 0,
        "not_found": 0,
        "successful": 0,
        "total": len(ingested_origins),
    }
    if with_listing:
        ingested_origins_table = []

    if status_counters["total"] == 0:
        exit(
            f"Forge {instance_name} has {len(ingested_origins)} \
scheduled ingest in the scheduler database."
        )

    for ingested_origin in ingested_origins:
        if ingested_origin.last_visit_status is not None:
            if with_listing:
                ingested_origins_table.append(
                    [
                        ingested_origin.url,
                        ingested_origin.last_visit_status.value,
                        str(ingested_origin.last_visit),
                    ]
                )
            counter_key = ingested_origin.last_visit_status.value
        else:
            counter_key = None
        status_counters[str(counter_key)] += 1

    return status_counters, ingested_origins_table if with_listing else []



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Agents using the new origin-aware scheduler."""

import logging
from typing import Any, Dict, Generator, Iterator, List

from simpy import Event

from swh.scheduler.journal_client import process_journal_objects

from .common import Environment, Queue, Task, TaskEvent

logger = logging.getLogger(__name__)


def scheduler_runner_process(
    env: Environment,
    task_queues: Dict[str, Queue],
    policy: str,
    min_batch_size: int,
) -> Iterator[Event]:
    """Scheduler runner. Grabs next visits from the database according to the
    scheduling policy, and fills the task_queues accordingly."""

    while True:
        for visit_type, queue in task_queues.items():
            remaining = queue.slots_remaining()
            if remaining < min_batch_size:
                continue
            next_origins = env.scheduler.grab_next_visits(
                visit_type, remaining, policy=policy, timestamp=env.time
            )
            logger.debug(
                "%s runner: running %s %s tasks",
                env.time,
                visit_type,
                len(next_origins),
            )
            for origin in next_origins:
                yield queue.put(Task(visit_type=origin.visit_type, origin=origin.url))

        yield env.timeout(10.0)


def scheduler_journal_client_process(
    env: Environment, status_queue: Queue
) -> Generator[Event, TaskEvent, None]:
    """Scheduler journal client. Every once in a while, pulls
    :class:`OriginVisitStatuses <swh.model.model.OriginVisitStatus>`
    from the status_queue to update the scheduler origin_visit_stats table."""
    BATCH_SIZE = 100

    statuses: List[Dict[str, Any]] = []
    while True:
        task_event = yield status_queue.get()
        statuses.append(task_event.status.to_dict())
        if len(statuses) < BATCH_SIZE:
            continue

        logger.debug(
            "%s journal client: processing %s statuses", env.time, len(statuses)
        )
        process_journal_objects(
            {"origin_visit_status": statuses}, scheduler=env.scheduler
        )

        statuses = []



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import json
import logging
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union
from uuid import UUID

import attr
from psycopg2.errors import CardinalityViolation
from psycopg2.extensions import AsIs
import psycopg2.extras
import psycopg2.pool
from testing.postgresql import Postgresql

from swh.core.db import BaseDb
from swh.core.db.common import db_transaction
from swh.core.db.db_utils import init_admin_extensions, populate_database_for_package
from swh.scheduler.model import (
    Task,
    TaskArguments,
    TaskPolicy,
    TaskPriority,
    TaskRun,
    TaskRunStatus,
    TaskStatus,
    TaskType,
)
from swh.scheduler.utils import utcnow

from .exc import SchedulerException, StaleData, UnknownPolicy
from .interface import ListedOriginPageToken, PaginatedListedOriginList
from .model import (
    LastVisitStatus,
    ListedOrigin,
    Lister,
    OriginVisitStats,
    SchedulerMetrics,
)

logger = logging.getLogger(__name__)


def adapt_LastVisitStatus(v: LastVisitStatus):
    return AsIs(f"'{v.value}'::last_visit_status")


psycopg2.extensions.register_adapter(dict, psycopg2.extras.Json)
psycopg2.extensions.register_adapter(LastVisitStatus, adapt_LastVisitStatus)
psycopg2.extras.register_uuid()


def format_query(query: str, keys: Sequence[str]) -> str:
    """Format a query with the given keys"""

    query_keys = ", ".join(keys)
    placeholders = ", ".join(["%s"] * len(keys))

    return query.format(keys=query_keys, placeholders=placeholders)


def mutate_task_dict(task_dict: Dict[str, Any]) -> Dict[str, Any]:
    task_dict["arguments"] = TaskArguments(**task_dict["arguments"])
    return task_dict


class SchedulerBackend:
    """Backend for the Software Heritage scheduling database."""

    current_version = 35

    def __init__(self, db, min_pool_conns=1, max_pool_conns=10):
        """
        Args:
            db_conn: either a libpq connection string, or a psycopg2 connection

        """
        if isinstance(db, psycopg2.extensions.connection):
            self._pool = None
            self._db = BaseDb(db)
        else:
            self._pool = psycopg2.pool.ThreadedConnectionPool(
                min_pool_conns,
                max_pool_conns,
                db,
                cursor_factory=psycopg2.extras.RealDictCursor,
            )
            self._db = None

    def get_db(self):
        if self._db:
            return self._db
        return BaseDb.from_pool(self._pool)

    def put_db(self, db):
        if db is not self._db:
            db.put_conn()

    task_type_keys = [field.name for field in attr.fields(TaskType)]

    @db_transaction()
    def create_task_type(self, task_type: TaskType, db=None, cur=None) -> None:
        """Create a new task type ready for scheduling.

        Args:
            task_type: a TaskType dictionary
        """
        query = format_query(
            """insert into task_type ({keys}) values ({placeholders})
            on conflict do nothing""",
            self.task_type_keys,
        )
        cur.execute(query, task_type.to_tuple())

    @db_transaction()
    def get_task_type(
        self, task_type_name: str, db=None, cur=None
    ) -> Optional[TaskType]:
        """Retrieve the task type with id task_type_name"""
        query = format_query(
            "select {keys} from task_type where type=%s",
            self.task_type_keys,
        )
        cur.execute(query, (task_type_name,))
        row = cur.fetchone()
        return TaskType(**row) if row is not None else None

    @db_transaction()
    def get_task_types(self, db=None, cur=None) -> List[TaskType]:
        """Retrieve all registered task types"""
        query = format_query(
            "select {keys} from task_type",
            self.task_type_keys,
        )
        cur.execute(query)
        return [TaskType(**row) for row in cur.fetchall()]

    @db_transaction()
    def get_listers(self, db=None, cur=None) -> List[Lister]:
        """Retrieve information about all listers from the database."""

        select_cols = ", ".join(Lister.select_columns())

        query = f"""
            select {select_cols} from listers
        """

        cur.execute(query)

        return [Lister(**ret) for ret in cur.fetchall()]

    @db_transaction()
    def get_listers_by_id(
        self, lister_ids: List[str], db=None, cur=None
    ) -> List[Lister]:
        """Retrieve listers in batch, using their UUID"""
        select_cols = ", ".join(Lister.select_columns())

        query = f"""
            select {select_cols} from listers
              where id in %s
        """

        if not lister_ids:
            return []

        cur.execute(query, (tuple(lister_ids),))

        return [Lister(**row) for row in cur]

    @db_transaction()
    def get_lister(
        self, name: str, instance_name: Optional[str] = None, db=None, cur=None
    ) -> Optional[Lister]:
        """Retrieve information about the given instance of the lister from the
        database.
        """
        if instance_name is None:
            instance_name = ""

        select_cols = ", ".join(Lister.select_columns())

        query = f"""
            select {select_cols} from listers
              where (name, instance_name) = (%s, %s)
        """

        cur.execute(query, (name, instance_name))

        ret = cur.fetchone()
        if not ret:
            return None

        return Lister(**ret)

    @db_transaction()
    def get_or_create_lister(
        self, name: str, instance_name: Optional[str] = None, db=None, cur=None
    ) -> Lister:
        """Retrieve information about the given instance of the lister from the
        database, or create the entry if it did not exist.
        """

        if instance_name is None:
            instance_name = ""

        select_cols = ", ".join(Lister.select_columns())
        insert_cols, insert_meta = (
            ", ".join(tup) for tup in Lister.insert_columns_and_metavars()
        )

        query = f"""
            with added as (
              insert into listers ({insert_cols}) values ({insert_meta})
                on conflict do nothing
                returning {select_cols}
            )
            select {select_cols} from added
          union all
            select {select_cols} from listers
              where (name, instance_name) = (%(name)s, %(instance_name)s);
        """

        cur.execute(query, Lister(name=name, instance_name=instance_name).to_dict())

        return Lister(**cur.fetchone())

    @db_transaction()
    def update_lister(self, lister: Lister, db=None, cur=None) -> Lister:
        """Update the state for the given lister instance in the database.

        Returns:
            a new Lister object, with all fields updated from the database

        Raises:
            StaleData if the `updated` timestamp for the lister instance in
        database doesn't match the one passed by the user.
        """

        select_cols = ", ".join(Lister.select_columns())
        set_vars = ", ".join(
            f"{col} = {meta}"
            for col, meta in zip(*Lister.insert_columns_and_metavars())
        )

        query = f"""update listers
                      set {set_vars}
                      where id=%(id)s and updated=%(updated)s
                      returning {select_cols}"""

        cur.execute(query, lister.to_dict())
        updated = cur.fetchone()

        if not updated:
            raise StaleData("Stale data; Lister state not updated")

        return Lister(**updated)

    @db_transaction()
    def record_listed_origins(
        self, listed_origins: Iterable[ListedOrigin], db=None, cur=None
    ) -> List[ListedOrigin]:
        """Record a set of origins that a lister has listed.

        This performs an "upsert": origins with the same (lister_id, url,
        visit_type) values are updated with new values for
        extra_loader_arguments, last_update and last_seen.
        """

        pk_cols = ListedOrigin.primary_key_columns()
        select_cols = ListedOrigin.select_columns()
        insert_cols, insert_meta = ListedOrigin.insert_columns_and_metavars()

        deduplicated_origins = {
            tuple(getattr(origin, k) for k in pk_cols): origin
            for origin in listed_origins
        }

        upsert_cols = [col for col in insert_cols if col not in pk_cols]
        upsert_set = ", ".join(f"{col} = EXCLUDED.{col}" for col in upsert_cols)

        query = f"""INSERT into listed_origins ({", ".join(insert_cols)})
                       VALUES %s
                    ON CONFLICT ({", ".join(pk_cols)}) DO UPDATE
                       SET {upsert_set}
                    RETURNING {", ".join(select_cols)}
        """

        ret = psycopg2.extras.execute_values(
            cur=cur,
            sql=query,
            argslist=(origin.to_dict() for origin in deduplicated_origins.values()),
            template=f"({', '.join(insert_meta)})",
            page_size=1000,
            fetch=True,
        )

        return [ListedOrigin(**d) for d in ret]

    @db_transaction()
    def get_listed_origins(
        self,
        lister_id: Optional[UUID] = None,
        url: Optional[str] = None,
        enabled: Optional[bool] = True,
        limit: int = 1000,
        page_token: Optional[ListedOriginPageToken] = None,
        db=None,
        cur=None,
    ) -> PaginatedListedOriginList:
        """Get information on the listed origins matching either the `url` or
        `lister_id`, or both arguments.
        """

        query_filters: List[str] = []
        query_params: List[Union[int, str, UUID, Tuple[UUID, str]]] = []

        if lister_id:
            query_filters.append("lister_id = %s")
            query_params.append(lister_id)

        if url is not None:
            query_filters.append("url = %s")
            query_params.append(url)

        if enabled is not None:
            query_filters.append("enabled = %s")
            query_params.append(enabled)

        if page_token is not None:
            query_filters.append("(lister_id, url) > %s")
            # the typeshed annotation for tuple() is too strict.
            query_params.append(tuple(page_token))  # type: ignore

        query_params.append(limit)

        select_cols = ", ".join(ListedOrigin.select_columns())
        if query_filters:
            where_clause = "where %s" % (" and ".join(query_filters))
        else:
            where_clause = ""

        query = f"""SELECT {select_cols}
                    from listed_origins
                    {where_clause}
                    ORDER BY lister_id, url
                    LIMIT %s"""

        cur.execute(query, tuple(query_params))
        origins = [ListedOrigin(**d) for d in cur]

        if len(origins) == limit:
            page_token = (str(origins[-1].lister_id), origins[-1].url)
        else:
            page_token = None

        return PaginatedListedOriginList(origins, page_token)

    @db_transaction()
    def grab_next_visits(
        self,
        visit_type: str,
        count: int,
        policy: str,
        enabled: bool = True,
        lister_uuid: Optional[str] = None,
        lister_name: Optional[str] = None,
        lister_instance_name: Optional[str] = None,
        timestamp: Optional[datetime.datetime] = None,
        absolute_cooldown: Optional[datetime.timedelta] = datetime.timedelta(hours=12),
        scheduled_cooldown: Optional[datetime.timedelta] = datetime.timedelta(days=7),
        failed_cooldown: Optional[datetime.timedelta] = datetime.timedelta(days=14),
        not_found_cooldown: Optional[datetime.timedelta] = datetime.timedelta(days=31),
        tablesample: Optional[float] = None,
        db=None,
        cur=None,
    ) -> List[ListedOrigin]:
        if timestamp is None:
            timestamp = utcnow()

        origin_select_cols = ", ".join(ListedOrigin.select_columns())

        joins: Dict[str, str] = {
            "origin_visit_stats": "USING (url, visit_type)",
        }

        query_args: List[Any] = []

        where_clauses = []

        # list of (name, query) handled as CTEs before the main query
        common_table_expressions: List[Tuple[str, str]] = []

        # "NOT enabled" = the lister said the origin no longer exists
        where_clauses.append("enabled" if enabled else "not enabled")

        # Only schedule visits of the given type
        where_clauses.append("visit_type = %s")
        query_args.append(visit_type)

        if absolute_cooldown:
            # Don't schedule visits if they've been scheduled since the absolute cooldown
            where_clauses.append(
                """origin_visit_stats.last_scheduled IS NULL
                   OR origin_visit_stats.last_scheduled < %s
                """
            )
            query_args.append(timestamp - absolute_cooldown)

        if scheduled_cooldown:
            # Don't re-schedule visits if they're already scheduled but we haven't
            # recorded a result yet, unless they've been scheduled more than a week
            # ago (it probably means we've lost them in flight somewhere).
            where_clauses.append(
                """origin_visit_stats.last_scheduled IS NULL
                OR origin_visit_stats.last_scheduled < GREATEST(
                  %s,
                  origin_visit_stats.last_visit
                )
            """
            )
            query_args.append(timestamp - scheduled_cooldown)

        if failed_cooldown:
            # Don't retry failed origins too often
            where_clauses.append(
                "origin_visit_stats.last_visit_status is distinct from 'failed' "
                "or origin_visit_stats.last_visit < %s"
            )
            query_args.append(timestamp - failed_cooldown)

        if not_found_cooldown:
            # Don't retry not found origins too often
            where_clauses.append(
                "origin_visit_stats.last_visit_status is distinct from 'not_found' "
                "or origin_visit_stats.last_visit < %s"
            )
            query_args.append(timestamp - not_found_cooldown)

        if policy == "oldest_scheduled_first":
            order_by = "origin_visit_stats.last_scheduled NULLS FIRST"
        elif policy == "never_visited_oldest_update_first":
            # never visited origins have a NULL last_snapshot
            where_clauses.append("origin_visit_stats.last_snapshot IS NULL")

            # order by increasing last_update (oldest first)
            where_clauses.append("listed_origins.last_update IS NOT NULL")
            order_by = "listed_origins.last_update"
        elif policy == "already_visited_order_by_lag":
            # TODO: store "visit lag" in a materialized view?

            # visited origins have a NOT NULL last_snapshot
            where_clauses.append("origin_visit_stats.last_snapshot IS NOT NULL")

            # ignore origins we have visited after the known last update
            where_clauses.append("listed_origins.last_update IS NOT NULL")
            where_clauses.append(
                "listed_origins.last_update > origin_visit_stats.last_successful"
            )

            # order by decreasing visit lag
            order_by = (
                "listed_origins.last_update - origin_visit_stats.last_successful DESC"
            )
        elif policy == "origins_without_last_update":
            where_clauses.append("last_update IS NULL")
            order_by = ", ".join(
                [
                    # By default, sort using the queue position. If the queue
                    # position is null, then the origin has never been visited,
                    # which we want to handle first
                    "origin_visit_stats.next_visit_queue_position nulls first",
                    # Schedule unknown origins in the order we've seen them
                    "listed_origins.first_seen",
                ]
            )

            # fmt: off

            # This policy requires updating the global queue position for this
            # visit type
            common_table_expressions.append(("update_queue_position", """
                INSERT INTO
                  visit_scheduler_queue_position(visit_type, position)
                SELECT
                  visit_type, COALESCE(MAX(next_visit_queue_position), 0)
                FROM selected_origins
                GROUP BY visit_type
                ON CONFLICT(visit_type) DO UPDATE
                  SET position=GREATEST(
                    visit_scheduler_queue_position.position, EXCLUDED.position
                  )
            """))
            # fmt: on
        else:
            raise UnknownPolicy(f"Unknown scheduling policy {policy}")

        if tablesample:
            table = "listed_origins tablesample SYSTEM (%s)"
            query_args.insert(0, tablesample)
        else:
            table = "listed_origins"

        if lister_uuid:
            where_clauses.append("lister_id = %s")
            query_args.append(lister_uuid)

        if lister_name:
            joins["listers"] = "on listed_origins.lister_id=listers.id"
            where_clauses.append("listers.name = %s")
            query_args.append(lister_name)

        if lister_instance_name:
            joins["listers"] = "on listed_origins.lister_id=listers.id"
            where_clauses.append("listers.instance_name = %s")
            query_args.append(lister_instance_name)

        join_clause = "\n".join(
            f"left join {table} {clause}" for table, clause in joins.items()
        )

        # fmt: off
        common_table_expressions.insert(0, ("selected_origins", f"""
            SELECT
              {origin_select_cols}, next_visit_queue_position
            FROM
              {table}
            {join_clause}
            WHERE
              ({") AND (".join(where_clauses)})
            ORDER BY
              {order_by}
            LIMIT %s
        """))
        # fmt: on

        query_args.append(count)

        # fmt: off
        common_table_expressions.append(("deduplicated_selected_origins", """
          SELECT DISTINCT
            url, visit_type
          FROM
            selected_origins
        """))
        # fmt: on

        # fmt: off
        common_table_expressions.append(("update_stats", """
            INSERT INTO
              origin_visit_stats (url, visit_type, last_scheduled)
            SELECT
              url, visit_type, %s
            FROM
              deduplicated_selected_origins
            ON CONFLICT (url, visit_type) DO UPDATE
              SET last_scheduled = GREATEST(
                origin_visit_stats.last_scheduled,
                EXCLUDED.last_scheduled
              )
        """))
        # fmt: on

        query_args.append(timestamp)

        formatted_ctes = ",\n".join(
            f"{name} AS (\n{cte}\n)" for name, cte in common_table_expressions
        )

        query = f"""
            WITH
              {formatted_ctes}
            SELECT
              {origin_select_cols}
            FROM
              selected_origins
        """

        cur.execute(query, tuple(query_args))
        return [ListedOrigin(**d) for d in cur]

    task_keys = [field.name for field in attr.fields(Task)]
    task_create_keys = [
        key for key in task_keys if key not in {"id", "current_interval"}
    ]

    @db_transaction()
    def create_tasks(
        self, tasks: List[Task], policy: TaskPolicy = "recurring", db=None, cur=None
    ) -> List[Task]:
        """Create new tasks.

        Args:
            tasks: each task is a Task object created with at least the following parameters:

                - type
                - arguments
                - next_run

            policy: default task policy (either recurring or oneshot) to use if not
                set in input task objects

        Returns:
            a list of created tasks with database ids filled

        """
        cur.execute("select swh_scheduler_mktemp_task()")
        db.copy_to(
            (task.to_dict() for task in tasks),
            "tmp_task",
            self.task_create_keys,
            default_values={"policy": policy},
            cur=cur,
        )
        query = format_query(
            "select {keys} from swh_scheduler_create_tasks_from_temp()",
            self.task_keys,
        )
        cur.execute(query)
        return [Task(**mutate_task_dict(row)) for row in cur.fetchall()]

    @db_transaction()
    def set_status_tasks(
        self,
        task_ids: List[int],
        status: TaskStatus = "disabled",
        next_run: Optional[datetime.datetime] = None,
        db=None,
        cur=None,
    ) -> None:
        """Set the tasks' status whose ids are listed.

        Args:
            task_ids: list of tasks' identifiers
            status: the status to set for the tasks
            next_run: if provided, also set the next_run date

        """
        if not task_ids:
            return
        query = ["UPDATE task SET status = %s"]
        args: List[Any] = [status]
        if next_run:
            query.append(", next_run = %s")
            args.append(next_run)
        query.append(" WHERE id IN %s")
        args.append(tuple(task_ids))

        cur.execute("".join(query), args)

    @db_transaction()
    def disable_tasks(self, task_ids: List[int], db=None, cur=None) -> None:
        """Disable the tasks whose ids are listed.

        Args:
            task_ids: list of tasks' identifiers
        """
        self.set_status_tasks(task_ids, db=db, cur=cur)

    @db_transaction()
    def search_tasks(
        self,
        task_id: Optional[int] = None,
        task_type: Optional[str] = None,
        status: Optional[TaskStatus] = None,
        priority: Optional[TaskPriority] = None,
        policy: Optional[TaskPolicy] = None,
        before: Optional[datetime.datetime] = None,
        after: Optional[datetime.datetime] = None,
        limit: Optional[int] = None,
        db=None,
        cur=None,
    ) -> List[Task]:
        """Search tasks from selected criterions

        Args:
            task_id: search a task with given identifier
            task_type: search tasks with given type
            status: search tasks with given status
            priority: search tasks with given priority
            policy: search tasks with given policy
            before: search tasks created before given date
            after: search tasks created after given date
            limit: maximum number of tasks to return

        Returns:
            a list of found tasks
        """
        where = []
        args: List[Any] = []

        if task_id:
            if isinstance(task_id, (str, int)):
                where.append("id = %s")
            else:
                where.append("id in %s")
                task_id = tuple(task_id)
            args.append(task_id)
        if task_type:
            if isinstance(task_type, str):
                where.append("type = %s")
            else:
                where.append("type in %s")
                task_type = tuple(task_type)
            args.append(task_type)
        if status:
            if isinstance(status, str):
                where.append("status = %s")
            else:
                where.append("status in %s")
                status = tuple(status)
            args.append(status)
        if priority:
            if isinstance(priority, str):
                where.append("priority = %s")
            else:
                priority = tuple(priority)
                where.append("priority in %s")
            args.append(priority)
        if policy:
            where.append("policy = %s")
            args.append(policy)
        if before:
            where.append("next_run <= %s")
            args.append(before)
        if after:
            where.append("next_run >= %s")
            args.append(after)

        query = "select * from task"
        if where:
            query += " where " + " and ".join(where)
        if limit:
            query += " limit %s :: bigint"
            args.append(limit)
        cur.execute(query, args)
        return [Task(**mutate_task_dict(row)) for row in cur.fetchall()]

    @db_transaction()
    def get_tasks(self, task_ids: List[int], db=None, cur=None) -> List[Task]:
        """Retrieve the info of tasks whose ids are listed.

        Args:
            task_ids: list of tasks' identifiers

        Returns:
            a list of Task objects
        """
        query = format_query("select {keys} from task where id in %s", self.task_keys)
        cur.execute(query, (tuple(task_ids),))
        return [Task(**mutate_task_dict(row)) for row in cur.fetchall()]

    @db_transaction()
    def peek_ready_tasks(
        self,
        task_type: str,
        timestamp: Optional[datetime.datetime] = None,
        num_tasks: Optional[int] = None,
        db=None,
        cur=None,
    ) -> List[Task]:
        """Fetch the list of tasks (with no priority) to be scheduled.

        Args:
            task_type: filtering task per their type
            timestamp: peek tasks that need to be executed
                before that timestamp
            num_tasks: only peek at num_tasks tasks (with no priority)

        Returns:
            the list of tasks which would be scheduled

        """
        if timestamp is None:
            timestamp = utcnow()

        cur.execute(
            """select * from swh_scheduler_peek_no_priority_tasks(
                %s, %s, %s :: bigint)""",
            (task_type, timestamp, num_tasks),
        )
        logger.debug("PEEK %s => %s" % (task_type, cur.rowcount))
        return [Task(**mutate_task_dict(row)) for row in cur.fetchall()]

    @db_transaction()
    def grab_ready_tasks(
        self,
        task_type: str,
        timestamp: Optional[datetime.datetime] = None,
        num_tasks: Optional[int] = None,
        db=None,
        cur=None,
    ) -> List[Task]:
        """Fetch and schedule the list of tasks (with no priority) ready to be scheduled.

        Args:
            task_type: filtering task per their type
            timestamp: grab tasks that need to be executed
                before that timestamp
            num_tasks: only grab num_tasks tasks (with no priority)

        Returns:
            the list of scheduled tasks

        """
        if timestamp is None:
            timestamp = utcnow()
        cur.execute(
            """select * from swh_scheduler_grab_ready_tasks(
                 %s, %s, %s :: bigint)""",
            (task_type, timestamp, num_tasks),
        )
        logger.debug("GRAB %s => %s" % (task_type, cur.rowcount))
        return [Task(**mutate_task_dict(row)) for row in cur.fetchall()]

    @db_transaction()
    def peek_ready_priority_tasks(
        self,
        task_type: str,
        timestamp: Optional[datetime.datetime] = None,
        num_tasks: Optional[int] = None,
        db=None,
        cur=None,
    ) -> List[Task]:
        """Fetch list of tasks (with any priority) ready to be scheduled.

        Args:
            task_type: filtering task per their type
            timestamp: peek tasks that need to be executed before that timestamp
            num_tasks: only peek at num_tasks tasks (with no priority)

        Returns:
            the list of tasks which would be scheduled

        """
        if timestamp is None:
            timestamp = utcnow()

        cur.execute(
            """select * from swh_scheduler_peek_any_ready_priority_tasks(
                %s, %s, %s :: bigint)""",
            (task_type, timestamp, num_tasks),
        )
        logger.debug("PEEK %s => %s", task_type, cur.rowcount)
        return [Task(**mutate_task_dict(row)) for row in cur.fetchall()]

    @db_transaction()
    def grab_ready_priority_tasks(
        self,
        task_type: str,
        timestamp: Optional[datetime.datetime] = None,
        num_tasks: Optional[int] = None,
        db=None,
        cur=None,
    ) -> List[Task]:
        """Fetch and schedule the list of tasks (with any priority) ready to be scheduled.

        Args:
            task_type: filtering task per their type
            timestamp: grab tasks that need to be executed
                before that timestamp
            num_tasks: only grab num_tasks tasks (with no priority)

        Returns:
            the list of scheduled tasks

        """
        if timestamp is None:
            timestamp = utcnow()
        cur.execute(
            """select * from swh_scheduler_grab_any_ready_priority_tasks(
                 %s, %s, %s :: bigint)""",
            (task_type, timestamp, num_tasks),
        )
        logger.debug("GRAB %s => %s", task_type, cur.rowcount)
        return [Task(**mutate_task_dict(row)) for row in cur.fetchall()]

    task_run_create_keys = ["task", "backend_id", "scheduled", "metadata"]

    @db_transaction()
    def schedule_task_run(
        self,
        task_id: int,
        backend_id: str,
        metadata: Optional[Dict[str, Any]] = None,
        timestamp: Optional[datetime.datetime] = None,
        db=None,
        cur=None,
    ) -> TaskRun:
        """Mark a given task as scheduled, adding a task_run entry in the database.

        Args:
            task_id: the identifier for the task being scheduled
            backend_id: the identifier of the job in the backend
            metadata: metadata to add to the task_run entry
            timestamp: the instant the event occurred

        Returns:
            a TaskRun object

        """

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = utcnow()

        cur.execute(
            "select * from swh_scheduler_schedule_task_run(%s, %s, %s, %s)",
            (task_id, backend_id, metadata, timestamp),
        )

        row = cur.fetchone()
        return TaskRun(**row)

    @db_transaction()
    def mass_schedule_task_runs(
        self, task_runs: List[TaskRun], db=None, cur=None
    ) -> None:
        """Schedule a bunch of task runs.

        Args:
            task_runs: a list of TaskRun objects created at least with the following parameters:

                - task
                - backend_id
                - scheduled
        """
        cur.execute("select swh_scheduler_mktemp_task_run()")
        db.copy_to(
            (task_run.to_dict() for task_run in task_runs),
            "tmp_task_run",
            self.task_run_create_keys,
            cur=cur,
        )
        cur.execute("select swh_scheduler_schedule_task_run_from_temp()")

    @db_transaction()
    def start_task_run(
        self,
        backend_id: str,
        metadata: Optional[Dict[str, Any]] = None,
        timestamp: Optional[datetime.datetime] = None,
        db=None,
        cur=None,
    ) -> Optional[TaskRun]:
        """Mark a given task as started, updating the corresponding task_run
           entry in the database.

        Args:
            backend_id: the identifier of the job in the backend
            metadata: metadata to add to the task_run entry
            timestamp: the instant the event occurred

        Returns:
            a TaskRun object with updated fields, or None if there was no
            TaskRun recorded with a matching backend_id.

        """

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = utcnow()

        cur.execute(
            "select * from swh_scheduler_start_task_run(%s, %s, %s)",
            (backend_id, metadata, timestamp),
        )

        for row in cur:
            if row["status"] is not None:
                return TaskRun(**row)

        logger.debug(
            "Failed to mark task run %s as started",
            backend_id,
        )
        return None

    @db_transaction()
    def end_task_run(
        self,
        backend_id: str,
        status: TaskRunStatus,
        metadata: Optional[Dict[str, Any]] = None,
        timestamp: Optional[datetime.datetime] = None,
        db=None,
        cur=None,
    ) -> Optional[TaskRun]:
        """Mark a given task as ended, updating the corresponding task_run entry in the
        database.

        Args:
            backend_id: the identifier of the job in the backend
            status: how the task ended
            metadata: metadata to add to the task_run entry
            timestamp: the instant the event occurred

        Returns:
            a TaskRun object with updated fields, or None if there was no
            TaskRun recorded with a matching backend_id.

        """

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = utcnow()

        cur.execute(
            "select * from swh_scheduler_end_task_run(%s, %s, %s, %s)",
            (backend_id, status, metadata, timestamp),
        )

        for row in cur:
            if row["status"] is not None:
                return TaskRun(**row)

        logger.debug(
            "Failed to mark task run %s as ended",
            backend_id,
        )
        return None

    @db_transaction()
    def filter_task_to_archive(
        self,
        after_ts: str,
        before_ts: str,
        limit: int = 10,
        page_token: Optional[str] = None,
        db=None,
        cur=None,
    ) -> Dict[str, Any]:
        """Compute the tasks to archive within the datetime interval
        [after_ts, before_ts[. The method returns a paginated result.

        Returns:
            dict with the following keys:
              - **next_page_token**: opaque token to be used as
                `page_token` to retrieve the next page of result. If absent,
                there is no more pages to gather.
              - **tasks**: list of task dictionaries with the following keys:

                    **id** (str): origin task id
                    **started** (Optional[datetime]): started date
                    **scheduled** (datetime): scheduled date
                    **arguments** (json dict): task's arguments
                    ...

        """
        assert not page_token or isinstance(page_token, str)
        last_id = -1 if page_token is None else int(page_token)
        tasks = []
        cur.execute(
            "select * from swh_scheduler_task_to_archive(%s, %s, %s, %s)",
            (after_ts, before_ts, last_id, limit + 1),
        )
        for row in cur:
            task = dict(row)
            # nested type index does not accept bare values
            # transform it as a dict to comply with this
            task["arguments"]["args"] = {
                i: v for i, v in enumerate(task["arguments"]["args"])
            }
            kwargs = task["arguments"]["kwargs"]
            task["arguments"]["kwargs"] = json.dumps(kwargs)
            tasks.append(task)

        if len(tasks) >= limit + 1:  # remains data, add pagination information
            result = {
                "tasks": tasks[:limit],
                "next_page_token": str(tasks[-1]["task_id"]),
            }
        else:
            result = {"tasks": tasks}

        return result

    @db_transaction()
    def delete_archived_tasks(self, task_ids, db=None, cur=None):
        """Delete archived tasks as much as possible. Only the task_ids whose
        complete associated task_run have been cleaned up will be.

        """
        _task_ids = _task_run_ids = []
        for task_id in task_ids:
            _task_ids.append(task_id["task_id"])
            _task_run_ids.append(task_id["task_run_id"])

        cur.execute(
            "select * from swh_scheduler_delete_archived_tasks(%s, %s)",
            (_task_ids, _task_run_ids),
        )

    task_run_keys = [
        "id",
        "task",
        "backend_id",
        "scheduled",
        "started",
        "ended",
        "metadata",
        "status",
    ]

    @db_transaction()
    def get_task_runs(
        self, task_ids: List[int], limit: Optional[int] = None, db=None, cur=None
    ) -> List[TaskRun]:
        """Search task run for a task id"""
        if task_ids:
            args: List[Any] = [tuple(task_ids)]
            query = "select * from task_run where task in %s"
            if limit:
                query += " limit %s :: bigint"
                args.append(limit)
            cur.execute(query, args)
            return [TaskRun(**row) for row in cur.fetchall()]
        else:
            return []

    @db_transaction()
    def origin_visit_stats_upsert(
        self, origin_visit_stats: Iterable[OriginVisitStats], db=None, cur=None
    ) -> None:
        pk_cols = OriginVisitStats.primary_key_columns()
        insert_cols, insert_meta = OriginVisitStats.insert_columns_and_metavars()

        upsert_cols = [col for col in insert_cols if col not in pk_cols]
        upsert_set = ", ".join(
            f"{col} = coalesce(EXCLUDED.{col}, ovi.{col})" for col in upsert_cols
        )

        query = f"""
            INSERT into origin_visit_stats AS ovi ({", ".join(insert_cols)})
            VALUES %s
            ON CONFLICT ({", ".join(pk_cols)}) DO UPDATE
            SET {upsert_set}
        """

        try:
            psycopg2.extras.execute_values(
                cur=cur,
                sql=query,
                argslist=(visit_stats.to_dict() for visit_stats in origin_visit_stats),
                template=f"({', '.join(insert_meta)})",
                page_size=1000,
                fetch=False,
            )
        except CardinalityViolation as e:
            raise SchedulerException(repr(e))

    @db_transaction()
    def origin_visit_stats_get(
        self, ids: Iterable[Tuple[str, str]], db=None, cur=None
    ) -> List[OriginVisitStats]:
        if not ids:
            return []
        primary_keys = tuple((origin, visit_type) for (origin, visit_type) in ids)
        query = format_query(
            """
            SELECT {keys}
            FROM (VALUES %s) as stats(url, visit_type)
            INNER JOIN origin_visit_stats USING (url, visit_type)
        """,
            OriginVisitStats.select_columns(),
        )
        rows = psycopg2.extras.execute_values(
            cur=cur, sql=query, argslist=primary_keys, fetch=True
        )
        return [OriginVisitStats(**row) for row in rows]

    @db_transaction()
    def visit_scheduler_queue_position_get(self, db=None, cur=None) -> Dict[str, int]:
        cur.execute("SELECT visit_type, position FROM visit_scheduler_queue_position")
        return {row["visit_type"]: row["position"] for row in cur}

    @db_transaction()
    def visit_scheduler_queue_position_set(
        self,
        visit_type: str,
        position: int,
        db=None,
        cur=None,
    ) -> None:
        query = """
            INSERT INTO visit_scheduler_queue_position(visit_type, position)
            VALUES(%s, %s)
            ON CONFLICT(visit_type) DO UPDATE SET position=EXCLUDED.position
        """
        cur.execute(query, (visit_type, position))

    @db_transaction()
    def update_metrics(
        self,
        lister_id: Optional[UUID] = None,
        timestamp: Optional[datetime.datetime] = None,
        db=None,
        cur=None,
    ) -> List[SchedulerMetrics]:
        """Update the performance metrics of this scheduler instance.

        Returns the updated metrics.

        Args:
          lister_id: if passed, update the metrics only for this lister instance
          timestamp: if passed, the date at which we're updating the metrics,
            defaults to the database NOW()
        """
        query = format_query(
            "SELECT {keys} FROM update_metrics(%s, %s)",
            SchedulerMetrics.select_columns(),
        )
        cur.execute(query, (lister_id, timestamp))
        return [SchedulerMetrics(**row) for row in cur.fetchall()]

    @db_transaction()
    def get_metrics(
        self,
        lister_id: Optional[UUID] = None,
        visit_type: Optional[str] = None,
        db=None,
        cur=None,
    ) -> List[SchedulerMetrics]:
        """Retrieve the performance metrics of this scheduler instance.

        Args:
          lister_id: filter the metrics for this lister instance only
          visit_type: filter the metrics for this visit type only
        """

        where_filters = []
        where_args = []
        if lister_id:
            where_filters.append("lister_id = %s")
            where_args.append(str(lister_id))
        if visit_type:
            where_filters.append("visit_type = %s")
            where_args.append(visit_type)

        where_clause = ""
        if where_filters:
            where_clause = f"where {' and '.join(where_filters)}"

        query = format_query(
            "SELECT {keys} FROM scheduler_metrics %s" % where_clause,
            SchedulerMetrics.select_columns(),
        )

        cur.execute(query, tuple(where_args))
        return [SchedulerMetrics(**row) for row in cur.fetchall()]


class TemporarySchedulerBackend(SchedulerBackend):
    """Temporary postgresql backend for the Software Heritage scheduling database.

    A temporary scheduler database is spawned then removed when the backend
    gets destroyed.

    It can be used for testing SWH components that require a scheduler instance
    (listers for instance).
    """

    def __init__(self):
        self.postgresql = Postgresql()
        init_admin_extensions("swh.scheduler", self.postgresql.url())
        populate_database_for_package("swh.scheduler", self.postgresql.url())
        super().__init__(self.postgresql.url())



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import csv
from dataclasses import dataclass, field
from datetime import datetime, timedelta
import textwrap
from typing import Dict, List, Optional, TextIO, Tuple
import uuid

import plotille
from simpy import Environment as _Environment
from simpy import Store

from swh.model.model import OriginVisitStatus
from swh.scheduler.interface import SchedulerInterface
from swh.scheduler.model import SchedulerMetrics


@dataclass
class SimulationReport:
    DURATION_THRESHOLD = 3600
    """Max duration for histograms"""

    total_visits: int = 0
    """Total count of finished visits"""

    visit_runtimes: Dict[Tuple[str, bool], List[float]] = field(default_factory=dict)
    """Collected visit runtimes for each (status, eventful) tuple"""

    scheduler_metrics: List[Tuple[datetime, List[SchedulerMetrics]]] = field(
        default_factory=list
    )
    """Collected scheduler metrics

    This is a list of couples (timestamp, [SchedulerMetrics,]): the list of
    scheduler metrics collected at given timestamp.
    """

    visit_metrics: List[Tuple[datetime, int]] = field(default_factory=list)
    """Collected visit metrics over time"""

    latest_snapshots: Dict[Tuple[str, str], bytes] = field(default_factory=dict)
    """Collected latest snapshots for origins"""

    def record_visit(
        self,
        origin: Tuple[str, str],
        duration: float,
        status: str,
        snapshot=Optional[bytes],
    ) -> None:
        eventful = False
        if status == "full":
            eventful = snapshot != self.latest_snapshots.get(origin)
            self.latest_snapshots[origin] = snapshot

        self.total_visits += 1
        self.visit_runtimes.setdefault((status, eventful), []).append(duration)

    def record_metrics(
        self, timestamp: datetime, scheduler_metrics: List[SchedulerMetrics]
    ):
        self.scheduler_metrics.append((timestamp, scheduler_metrics))
        self.visit_metrics.append((timestamp, self.total_visits))

    @property
    def uneventful_visits(self):
        """Number of uneventful, full visits"""
        return len(self.visit_runtimes.get(("full", False), []))

    def runtime_histogram(self, status: str, eventful: bool) -> str:
        runtimes = self.visit_runtimes.get((status, eventful), [])
        return plotille.hist(
            [runtime for runtime in runtimes if runtime <= self.DURATION_THRESHOLD]
        )

    def metrics_plot(self) -> str:
        timestamps, metric_lists = zip(*self.scheduler_metrics)
        known = [sum(m.origins_known for m in metrics) for metrics in metric_lists]
        never_visited = [
            sum(m.origins_never_visited for m in metrics) for metrics in metric_lists
        ]

        figure = plotille.Figure()
        figure.x_label = "simulated time"
        figure.y_label = "origins"
        figure.scatter(timestamps, known, label="Known origins")
        figure.scatter(timestamps, never_visited, label="Origins never visited")

        visit_timestamps, n_visits = zip(*self.visit_metrics)
        figure.scatter(visit_timestamps, n_visits, label="Visits over time")

        return figure.show(legend=True)

    def metrics_csv(self, fobj: TextIO) -> None:
        """Export scheduling metrics in a csv file"""
        csv_writer = csv.writer(fobj)
        csv_writer.writerow(
            [
                "timestamp",
                "known_origins",
                "enabled_origins",
                "never_visited_origins",
                "origins_with_pending_changes",
            ]
        )

        timestamps, metric_lists = zip(*self.scheduler_metrics)
        known = (sum(m.origins_known for m in metrics) for metrics in metric_lists)
        enabled = (sum(m.origins_enabled for m in metrics) for metrics in metric_lists)
        never_visited = (
            sum(m.origins_never_visited for m in metrics) for metrics in metric_lists
        )
        pending_changes = (
            sum(m.origins_with_pending_changes for m in metrics)
            for metrics in metric_lists
        )
        csv_writer.writerows(
            zip(timestamps, known, enabled, never_visited, pending_changes)
        )

    def format(self, with_plots=True):
        full_visits = self.visit_runtimes.get(("full", True), [])
        long_tasks = sum(runtime > self.DURATION_THRESHOLD for runtime in full_visits)

        output = textwrap.dedent(
            f"""\
                Total visits: {self.total_visits}
                Uneventful visits: {self.uneventful_visits}
                Eventful visits: {len(full_visits)}
                Very long running tasks: {long_tasks}
                """
        )
        if with_plots:
            histogram = self.runtime_histogram("full", True)
            plot = self.metrics_plot()
            output += (
                "Visit time histogram for eventful visits:"
                + histogram
                + "\n"
                + textwrap.dedent(
                    """\
                    Metrics over time:
                    """
                )
                + plot
            )
        return output


@dataclass
class Task:
    visit_type: str
    origin: str
    backend_id: uuid.UUID = field(default_factory=uuid.uuid4)


@dataclass
class TaskEvent:
    task: Task
    status: OriginVisitStatus
    eventful: bool = field(default=False)


class Queue(Store):
    """Model a queue of objects to be passed between processes."""

    def __len__(self):
        return len(self.items or [])

    def slots_remaining(self):
        return self.capacity - len(self)


class Environment(_Environment):
    report: SimulationReport
    scheduler: SchedulerInterface

    def __init__(self, start_time: datetime):
        if start_time.tzinfo is None:
            raise ValueError("start_time must have timezone information")
        self.start_time = start_time
        super().__init__()

    @property
    def time(self):
        """Get the current simulated wall clock time"""
        return self.start_time + timedelta(seconds=self.now)



# Copyright (C) 2017-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import datetime
from typing import Any, Dict, List, Optional

from swh.scheduler.model import Task, TaskArguments, TaskPriority, TaskType
from swh.scheduler.utils import utcnow

TEMPLATES = {
    "test-git": Task(
        type="load-test-git",
        arguments=TaskArguments(),
        next_run=utcnow(),
    ),
    "test-hg": Task(
        type="load-test-hg",
        arguments=TaskArguments(),
        next_run=utcnow(),
        policy="oneshot",
    ),
}


TASK_TYPES = {
    "test-git": TaskType(
        type="load-test-git",
        description="Update a git repository",
        backend_name="swh.loader.git.tasks.UpdateGitRepository",
        default_interval=datetime.timedelta(days=64),
        min_interval=datetime.timedelta(hours=12),
        max_interval=datetime.timedelta(days=64),
        backoff_factor=2.0,
        max_queue_length=None,
        num_retries=7,
        retry_delay=datetime.timedelta(hours=2),
    ),
    "test-hg": TaskType(
        type="load-test-hg",
        description="Update a mercurial repository",
        backend_name="swh.loader.mercurial.tasks.UpdateHgRepository",
        default_interval=datetime.timedelta(days=64),
        min_interval=datetime.timedelta(hours=12),
        max_interval=datetime.timedelta(days=64),
        backoff_factor=2.0,
        max_queue_length=None,
        num_retries=7,
        retry_delay=datetime.timedelta(hours=2),
    ),
}


def _task_from_template(
    template: Task,
    next_run: datetime.datetime,
    priority: Optional[TaskPriority],
    *args,
    **kwargs,
) -> Task:
    fields_to_update: Dict[str, Any] = {"next_run": next_run}
    if priority:
        fields_to_update["priority"] = priority
    if args or kwargs:
        fields_to_update["arguments"] = TaskArguments(
            args=list(args) or [], kwargs=kwargs or {}
        )
    return template.evolve(**fields_to_update)


def tasks_from_template(
    template: Task,
    max_timestamp: datetime.datetime,
    num: Optional[int] = None,
    priority: Optional[TaskPriority] = None,
    num_priorities: Dict[Optional[TaskPriority], int] = {},
) -> List[Task]:
    """Build ``num`` tasks from template"""
    assert bool(num) != bool(num_priorities), "mutually exclusive"
    if not num_priorities:
        assert num is not None  # to please mypy
        num_priorities = {None: num}
    tasks: List[Task] = []
    for priority, num in num_priorities.items():
        for _ in range(num):
            i = len(tasks)
            tasks.append(
                _task_from_template(
                    template,
                    max_timestamp - datetime.timedelta(microseconds=i),
                    priority,
                    "argument-%03d" % i,
                    **{"kwarg%03d" % i: "bogus-kwarg"},
                )
            )
    return tasks


def tasks_with_priority_from_template(
    template: Task, max_timestamp: datetime.datetime, num: int, priority: TaskPriority
) -> List[Task]:
    """Build tasks with priority from template"""
    return [
        _task_from_template(
            template,
            max_timestamp - datetime.timedelta(microseconds=i),
            priority,
            "argument-%03d" % i,
            **{"kwarg%03d" % i: "bogus-kwarg"},
        )
        for i in range(num)
    ]


LISTERS = (
    {"name": "github", "instance_name": "github"},
    {"name": "gitlab", "instance_name": "gitlab"},
    {"name": "gitlab", "instance_name": "freedesktop"},
    {"name": "npm"},
    {"name": "pypi"},
)



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import io

import pytest

from swh.core.api.classes import stream_results
import swh.scheduler.simulator as simulator
from swh.scheduler.tests.common import TASK_TYPES

NUM_ORIGINS = 42
TEST_RUNTIME = 1000


def test_fill_test_data(swh_scheduler):
    for task_type in TASK_TYPES.values():
        swh_scheduler.create_task_type(task_type)

    simulator.fill_test_data(swh_scheduler, num_origins=NUM_ORIGINS)

    origins = list(stream_results(swh_scheduler.get_listed_origins))
    assert len(origins) == NUM_ORIGINS

    res = swh_scheduler.search_tasks()
    assert len(res) == NUM_ORIGINS


@pytest.mark.parametrize(
    "policy",
    (
        "oldest_scheduled_first",
        "never_visited_oldest_update_first",
        "already_visited_order_by_lag",
    ),
)
def test_run_origin_scheduler(swh_scheduler, policy):
    for task_type in TASK_TYPES.values():
        swh_scheduler.create_task_type(task_type)

    simulator.fill_test_data(swh_scheduler, num_origins=NUM_ORIGINS)
    simulator.run(
        swh_scheduler,
        scheduler_type="origin_scheduler",
        policy=policy,
        runtime=TEST_RUNTIME,
    )


def test_run_task_scheduler(swh_scheduler):
    for task_type in TASK_TYPES.values():
        swh_scheduler.create_task_type(task_type)

    simulator.fill_test_data(swh_scheduler, num_origins=NUM_ORIGINS)
    report = simulator.run(
        swh_scheduler,
        scheduler_type="task_scheduler",
        policy=None,
        runtime=TEST_RUNTIME,
    )

    # just check these SimulationReport methods do not crash
    assert report.format(with_plots=True)
    assert report.format(with_plots=False)
    fobj = io.StringIO()
    report.metrics_csv(fobj=fobj)
    assert fobj.getvalue()



# Copyright (C) 2017-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Defines the <swh scheduler origin send-origins-from-file-to-celery> cli utility
functions. This uses a list of origins read from the standard input or file, massage
them into scheduler tasks to send directly to celery to a queue (according to a task
type specified).

The list of origins has been extracted by other means (e.g. sentry extract, combination
of various shell scripts, ...). Then, a human operator provides the list to the cli so
it's consumed by standard swh queues (understand scheduler configured backend).

"""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional

    from swh.scheduler.interface import SchedulerInterface
    from swh.scheduler.model import TaskType


def get_scheduler_task_type(
    scheduler: SchedulerInterface, task_type_name: str
) -> TaskType:
    """Retrieve a TaskType instance for a task type name from the scheduler.

    Args:
        scheduler: Scheduler instance to lookup data from
        task_type_name: The task type name to lookup

    Raises:
        ValueError when task_type_name or its fallback are not found.

    Returns:
        Information about the task type

    """
    origin_task_type_name = task_type_name
    # Lookup standard scheduler task (e.g. load-git, load-hg, ...)
    scheduler_info = scheduler.get_task_type(task_type_name)
    while True:
        if scheduler_info:
            return scheduler_info
        # Lookup task type derivative (e.g. load-git-large, load-hg-bitbucket, ...)
        new_task_type_name = task_type_name.rsplit("-", 1)[0]
        if new_task_type_name == task_type_name:
            error_msg = f"Could not find scheduler <{origin_task_type_name}> task type"
            raise ValueError(error_msg)
        task_type_name = new_task_type_name
        scheduler_info = scheduler.get_task_type(task_type_name)


def lines_to_task_args(
    lines: Iterable[str],
    columns: List[str] = ["url"],
    postprocess: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
    **kwargs,
) -> Iterator[Dict[str, Any]]:
    """Iterate over the lines and convert them into celery tasks ready to be sent.

    Args:
        lines: Line read from a file or stdin
        columns: structure of the lines to be read (usually only the url column)
        postprocess: An optional callable to enrich the task with
        **kwargs: extra static arguments to enrich the task with

    Yield:
        task ready to be sent to celery

    """
    for line in lines:
        values = line.strip().split()
        ret = dict(zip(columns, values))
        ret.update(kwargs)
        if postprocess:
            ret = postprocess(ret)
        yield {"args": [], "kwargs": ret}


_to_mercurial_tasks = partial(
    lines_to_task_args, columns=["origin_url", "archive_path"]
)

_to_bitbucket_mercurial_tasks = partial(
    lines_to_task_args, columns=["url", "directory", "visit_date"]
)


def _to_svn_tasks(lines: Iterable[str], type: str = "svn", **kwargs) -> Iterator[Dict]:
    """Generates proper task argument for the loader-svn worker.

    Yields:
        svn task

    """
    if type == "svn":
        yield from lines_to_task_args(lines=lines, columns=["url"], **kwargs)
    else:
        yield from lines_to_task_args(
            lines=lines,
            columns=["origin_url", "archive_path"],
            **kwargs,
        )


def _update_git_task_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Create simple git task kwargs from an url."""

    from urllib.parse import urlparse

    ret = kwargs.copy()

    parsed_url = urlparse(ret["url"])
    if parsed_url.netloc == "github.com":
        ret.update(  # extra information for the loader-metadata
            {
                "lister_name": "github",
                "lister_instance_name": "github",
            }
        )

    return ret


# Generates task argument for loader-git for usual origins
_to_git_normal_tasks = partial(
    lines_to_task_args,
    columns=["url"],
    postprocess=_update_git_task_kwargs,
)


# Generates task argument for loader-git for origins whose packfile is very large
_to_git_large_tasks = partial(
    _to_git_normal_tasks,
    pack_size_bytes=34359738368,
    verify_certs=False,
)

# Default task arguments generator for previously scheduled kind of origins
# It lifts the other callables into a simple dict to ease
TASK_ARGS_GENERATOR_CALLABLES: Dict[str, Callable] = {
    "load-svn": _to_svn_tasks,
    "load-hg-from-archive-mercurial": partial(
        _to_mercurial_tasks,
        visit_date="Tue, 3 May 2016 17:16:32 +0200",
    ),
    "load-hg-bitbucket": _to_bitbucket_mercurial_tasks,
    "load-git-normal": _to_git_normal_tasks,
    "load-git-large": _to_git_large_tasks,
}



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
import yaml

from swh.scheduler.api.server import load_and_check_config


def prepare_config_file(tmpdir, content, name="config.yml"):
    """Prepare configuration file in `$tmpdir/name` with content `content`.

    Args:
        tmpdir (LocalPath): root directory
        content (str/dict): Content of the file either as string or as a dict.
                            If a dict, converts the dict into a yaml string.
        name (str): configuration filename

    Returns
        path (str) of the configuration file prepared.

    """
    config_path = tmpdir / name
    if isinstance(content, dict):  # convert if needed
        content = yaml.dump(content)
    config_path.write_text(content, encoding="utf-8")
    # pytest on python3.5 does not support LocalPath manipulation, so
    # convert path to string
    return str(config_path)


@pytest.mark.parametrize("scheduler_class", [None, ""])
def test_load_and_check_config_no_configuration(scheduler_class):
    """Inexistent configuration files raises"""
    with pytest.raises(EnvironmentError, match="Configuration file must be defined"):
        load_and_check_config(scheduler_class)


def test_load_and_check_config_inexistent_fil():
    """Inexistent config filepath should raise"""
    config_path = "/some/inexistent/config.yml"
    expected_error = f"Configuration file {config_path} does not exist"
    with pytest.raises(FileNotFoundError, match=expected_error):
        load_and_check_config(config_path)


def test_load_and_check_config_wrong_configuration(tmpdir):
    """Wrong configuration raises"""
    config_path = prepare_config_file(tmpdir, "something: useless")
    with pytest.raises(KeyError, match="Missing '%scheduler' configuration"):
        load_and_check_config(config_path)


def test_load_and_check_config_remote_config_local_type_raise(tmpdir):
    """Configuration without 'postgresql' storage is rejected"""
    config_path = prepare_config_file(tmpdir, {"scheduler": {"cls": "remote"}})
    with pytest.raises(ValueError, match="'postgresql'"):
        load_and_check_config(config_path)


def test_load_and_check_config_local_incomplete_configuration(tmpdir):
    """Incomplete 'local' configuration should raise"""
    config = {
        "scheduler": {
            "cls": "postgresql",
            "something": "needed-for-test",
        }
    }

    config_path = prepare_config_file(tmpdir, config)
    expected_error = "Invalid configuration; missing 'db' config entry"
    with pytest.raises(KeyError, match=expected_error):
        load_and_check_config(config_path)


@pytest.mark.parametrize("clazz", ["local", "postgresql"])
def test_load_and_check_config_local_config_fine(tmpdir, clazz):
    """Local configuration is fine"""
    config = {
        "scheduler": {
            "cls": clazz,
            "db": "db",
        }
    }
    config_path = prepare_config_file(tmpdir, config)
    cfg = load_and_check_config(config_path, type="postgresql")
    assert cfg == config


def test_load_and_check_config_remote_config_fine(tmpdir):
    """Remote configuration is fine"""
    config = {"scheduler": {"cls": "remote"}}
    config_path = prepare_config_file(tmpdir, config)
    cfg = load_and_check_config(config_path, type="any")
    assert cfg == config



# Copyright (C) 2020-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import attr

from swh.scheduler import model


def test_select_columns():
    @attr.s
    class TestModel(model.BaseSchedulerModel):
        id = attr.ib(type=str)
        test1 = attr.ib(type=str)
        a_first_attr = attr.ib(type=str)

        @property
        def test2(self):
            """This property should not show up in the extracted columns"""
            return self.test1

    assert TestModel.select_columns() == ("a_first_attr", "id", "test1")


def test_insert_columns():
    @attr.s
    class TestModel(model.BaseSchedulerModel):
        id = attr.ib(type=str)
        test1 = attr.ib(type=str)

        @property
        def test2(self):
            """This property should not show up in the extracted columns"""
            return self.test1

    assert TestModel.insert_columns_and_metavars() == (
        ("id", "test1"),
        ("%(id)s", "%(test1)s"),
    )


def test_insert_columns_auto_now_add():
    @attr.s
    class TestModel(model.BaseSchedulerModel):
        id = attr.ib(type=str)
        test1 = attr.ib(type=str)
        added = attr.ib(type=datetime.datetime, metadata={"auto_now_add": True})

    assert TestModel.insert_columns_and_metavars() == (
        ("id", "test1"),
        ("%(id)s", "%(test1)s"),
    )


def test_insert_columns_auto_now():
    @attr.s
    class TestModel(model.BaseSchedulerModel):
        id = attr.ib(type=str)
        test1 = attr.ib(type=str)
        updated = attr.ib(type=datetime.datetime, metadata={"auto_now": True})

    assert TestModel.insert_columns_and_metavars() == (
        ("id", "test1", "updated"),
        ("%(id)s", "%(test1)s", "now()"),
    )


def test_insert_columns_primary_key():
    @attr.s
    class TestModel(model.BaseSchedulerModel):
        id = attr.ib(type=str, metadata={"auto_primary_key": True})
        test1 = attr.ib(type=str)

    assert TestModel.insert_columns_and_metavars() == (("test1",), ("%(test1)s",))


def test_insert_primary_key():
    @attr.s
    class TestModel(model.BaseSchedulerModel):
        id = attr.ib(type=str, metadata={"auto_primary_key": True})
        test1 = attr.ib(type=str)

    assert TestModel.primary_key_columns() == ("id",)

    @attr.s
    class TestModel2(model.BaseSchedulerModel):
        col1 = attr.ib(type=str, metadata={"primary_key": True})
        col2 = attr.ib(type=str, metadata={"primary_key": True})
        test1 = attr.ib(type=str)

    assert TestModel2.primary_key_columns() == ("col1", "col2")



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Agents using the "old" task-based scheduler."""

import logging
from typing import Dict, Generator, Iterator

from simpy import Event

from swh.scheduler.model import TaskRun, TaskRunStatus

from .common import Environment, Queue, Task, TaskEvent

logger = logging.getLogger(__name__)


def scheduler_runner_process(
    env: Environment,
    task_queues: Dict[str, Queue],
    min_batch_size: int,
) -> Iterator[Event]:
    """Scheduler runner. Grabs next visits from the database according to the
    scheduling policy, and fills the task_queues accordingly."""

    while True:
        for visit_type, queue in task_queues.items():
            remaining = queue.slots_remaining()
            if remaining < min_batch_size:
                continue
            next_tasks = env.scheduler.grab_ready_tasks(
                f"load-{visit_type}", num_tasks=remaining, timestamp=env.time
            )
            logger.debug(
                "%s runner: running %s %s tasks",
                env.time,
                visit_type,
                len(next_tasks),
            )

            sim_tasks = [
                Task(visit_type=visit_type, origin=task.arguments.kwargs["url"])
                for task in next_tasks
            ]

            env.scheduler.mass_schedule_task_runs(
                [
                    TaskRun(
                        task=task.id,
                        scheduled=env.time,
                        backend_id=str(sim_task.backend_id),
                    )
                    for task, sim_task in zip(next_tasks, sim_tasks)
                ]
            )

            for sim_task in sim_tasks:
                yield queue.put(sim_task)

        yield env.timeout(10.0)


def scheduler_listener_process(
    env: Environment, status_queue: Queue
) -> Generator[Event, TaskEvent, None]:
    """Scheduler listener. In the real world this would listen to celery
    events, but we listen to the status_queue and simulate celery events from
    that."""
    while True:
        event = yield status_queue.get()
        if event.status.status == "ongoing":
            env.scheduler.start_task_run(str(event.task.backend_id), timestamp=env.time)
        else:
            status: TaskRunStatus = "failed"
            if event.status.status == "full":
                status = "eventful" if event.eventful else "uneventful"

            env.scheduler.end_task_run(
                str(event.task.backend_id), status=status, timestamp=env.time
            )



# Copyright (C) 2015-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
import logging
import os
import sys
from time import monotonic as _monotonic
import traceback
from typing import Any, Dict, Optional
import urllib.parse

from celery import Celery
from celery.signals import celeryd_after_setup, setup_logging, task_prerun, worker_init
from celery.utils.log import ColorFormatter
from celery.worker.control import Panel
from kombu import Exchange, Queue
import pkg_resources
import requests

from swh.core.config import load_named_config, merge_configs
from swh.core.sentry import init_sentry
from swh.scheduler import CONFIG as SWH_CONFIG

DEFAULT_CONFIG_NAME = "worker"
CONFIG_NAME_ENVVAR = "SWH_WORKER_INSTANCE"
CONFIG_NAME_TEMPLATE = "worker/%s"

DEFAULT_CONFIG = {
    "task_broker": ("str", "amqp://guest@localhost//"),
    "task_modules": ("list[str]", []),
    "task_queues": ("list[str]", []),
    "task_soft_time_limit": ("int", 0),
}

logger = logging.getLogger(__name__)


# Celery eats tracebacks in signal callbacks, this decorator catches
# and prints them.
# Also tries to notify Sentry if possible.
def _print_errors(f):
    @functools.wraps(f)
    def newf(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as exc:
            traceback.print_exc()
            try:
                import sentry_sdk

                sentry_sdk.capture_exception(exc)
            except Exception:
                traceback.print_exc()

    return newf


@setup_logging.connect
@_print_errors
def setup_log_handler(
    loglevel=None,
    logfile=None,
    format=None,
    colorize=None,
    log_console=None,
    log_journal=None,
    **kwargs,
):
    """Setup logging according to Software Heritage preferences.

    If the environment variable SWH_LOG_CONFIG is provided, this uses the targeted
    logging configuration file to configure logging. Otherwise, as before, this uses the
    default enclosed coded configuration.

    """
    from swh.core.logging import logging_configure

    # Retrieve logger configuration yaml filepath from environment variable if any
    log_config_path = os.environ.get("SWH_LOG_CONFIG")
    if log_config_path is not None:
        # Delegate configuration to the log_config_path
        logging_configure([], log_config_path)
    else:
        # Keep the logging config coming from this code block
        if loglevel is None:
            loglevel = logging.DEBUG
        if isinstance(loglevel, str):
            loglevel = logging.getLevelName(loglevel)

        formatter = logging.Formatter(format)

        root_logger = logging.getLogger("")
        root_logger.setLevel(logging.INFO)

        log_target = os.environ.get("SWH_LOG_TARGET", "console")
        if log_target == "console":
            log_console = True
        elif log_target == "journal":
            log_journal = True

        # this looks for log levels *higher* than DEBUG
        if loglevel <= logging.DEBUG and log_console is None:
            log_console = True

        if log_console:
            color_formatter = ColorFormatter(format) if colorize else formatter
            console = logging.StreamHandler()
            console.setLevel(logging.DEBUG)
            console.setFormatter(color_formatter)
            root_logger.addHandler(console)

        if log_journal:
            try:
                from swh.core.logger import JournalHandler

                systemd_journal = JournalHandler()
                systemd_journal.setLevel(logging.DEBUG)
                systemd_journal.setFormatter(formatter)
                root_logger.addHandler(systemd_journal)
            except ImportError:
                root_logger.warning(
                    "JournalHandler is not available, skipping. "
                    "Please install swh-core[logging]."
                )

        # Historical configuration kept as-is
        logging_configure(
            [
                ("celery", logging.INFO),
                # Silence amqp heartbeat_tick messages
                ("amqp", loglevel),
                # Silence useless "Starting new HTTP connection" messages
                ("urllib3", logging.WARNING),
                # Completely disable azure logspam
                ("azure.core.pipeline.policies.http_logging_policy", logging.WARNING),
                ("swh", loglevel),
                # get_task_logger makes the swh tasks loggers children of celery.task
                ("celery.task", loglevel),
            ]
        )

    # extra step for amqp
    logger = logging.getLogger("amqp")
    logger.addFilter(lambda record: not record.msg.startswith("heartbeat_tick"))
    return loglevel


@celeryd_after_setup.connect
@_print_errors
def setup_queues_and_tasks(sender, instance, **kwargs):
    """Signal called on worker start.

    This automatically registers swh.scheduler.task.Task subclasses as
    available celery tasks.

    This also subscribes the worker to the "implicit" per-task queues defined
    for these task classes.

    """
    logger.info("Setup Queues & Tasks for %s", sender)
    instance.app.conf["worker_name"] = sender


def _init_sentry(sentry_dsn: Optional[str] = None, main_package: Optional[str] = None):
    try:
        from sentry_sdk.integrations.celery import CeleryIntegration
    except ImportError:
        integrations = []
    else:
        integrations = [CeleryIntegration()]
    init_sentry(
        sentry_dsn,
        integrations=integrations,
        main_package=main_package,
        deferred_init=sentry_dsn is None,
    )


@worker_init.connect
@_print_errors
def on_worker_init(*args, **kwargs):
    # init sentry with no DSN first to ensure celery integration for sentry is
    # properly configured as it must happen in the worker_init signal callback,
    # real sentry DSN is then setup in task_prerun signal callback
    # (see celery_task_prerun function below)
    _init_sentry()

    if "pytest" in sys.argv[0] or "PYTEST_XDIST_WORKER" in os.environ:
        # when pytest collects tests, it breaks the proper configuration
        # of the celery integration as a side effect, so we ensure that
        # the celery.worker.consumer.build_tracer function gets overridden
        # as it should have be
        from celery.app import trace
        from celery.worker.consumer import consumer

        consumer.build_tracer = trace.build_tracer


@Panel.register
def monotonic(state):
    """Get the current value for the monotonic clock"""
    return {"monotonic": _monotonic()}


def route_for_task(name, args, kwargs, options, task=None, **kw):
    """Route tasks according to the task_queue attribute in the task class"""
    if name is not None and name.startswith("swh."):
        return {"queue": name}


def get_queue_stats(app, queue_name):
    """Get the statistics regarding a queue on the broker.

    Arguments:
      queue_name: name of the queue to check

    Returns a dictionary raw from the RabbitMQ management API;
    or `None` if the current configuration does not use RabbitMQ.

    Interesting keys:
     - Consumers (number of consumers for the queue)
     - messages (number of messages in queue)
     - messages_unacknowledged (number of messages currently being
       processed)

    Documentation: https://www.rabbitmq.com/management.html#http-api
    """

    conn_info = app.connection().info()
    if conn_info["transport"] == "memory":
        # We're running in a test environment, without RabbitMQ.
        return None
    url = "http://{hostname}:{port}/api/queues/{vhost}/{queue}".format(
        hostname=conn_info["hostname"],
        port=conn_info["port"] + 10000,
        vhost=urllib.parse.quote(conn_info["virtual_host"], safe=""),
        queue=urllib.parse.quote(queue_name, safe=""),
    )
    credentials = (conn_info["userid"], conn_info["password"])
    r = requests.get(url, auth=credentials)
    if r.status_code == 404:
        return {}
    if r.status_code != 200:
        raise ValueError(
            "Got error %s when reading queue stats: %s" % (r.status_code, r.json())
        )
    return r.json()


def get_queue_length(app, queue_name):
    """Shortcut to get a queue's length"""
    stats = get_queue_stats(app, queue_name)
    if stats:
        return stats.get("messages")


MAX_NUM_TASKS = 10000


def get_available_slots(app, queue_name: str, max_length: Optional[int]):
    """Get the number of tasks that can be sent to `queue_name`, when
    the queue is limited to `max_length`.

    Returns:
        The number of available slots in the queue. That result should be positive.

    """

    if not max_length:
        return MAX_NUM_TASKS

    try:
        queue_length = get_queue_length(app, queue_name)
        # Clamp the return value to MAX_NUM_TASKS
        max_val = max(0, min(max_length - queue_length, MAX_NUM_TASKS))
    except (ValueError, TypeError):
        # Unknown queue length, just schedule all the tasks
        max_val = max_length

    return max_val


def register_task_class(app, name, cls):
    """Register a class-based task under the given name"""
    if name in app.tasks:
        return

    task_instance = cls()
    task_instance.name = name
    app.register_task(task_instance)


INSTANCE_NAME = os.environ.get(CONFIG_NAME_ENVVAR)
CONFIG_NAME = os.environ.get("SWH_CONFIG_FILENAME")
CONFIG: Dict[str, Any] = {}

if CONFIG_NAME:
    # load the celery config from the main config file given as
    # SWH_CONFIG_FILENAME environment variable.
    # This is expected to have a [celery] section in which we have the
    # celery specific configuration.
    SWH_CONFIG.clear()
    SWH_CONFIG.update(load_named_config(CONFIG_NAME))
    CONFIG = SWH_CONFIG.get("celery", {})

if not CONFIG:
    # otherwise, back to compat config loading mechanism
    if INSTANCE_NAME:
        CONFIG_NAME = CONFIG_NAME_TEMPLATE % INSTANCE_NAME
    else:
        CONFIG_NAME = DEFAULT_CONFIG_NAME

    # Load the Celery config
    CONFIG = load_named_config(CONFIG_NAME, DEFAULT_CONFIG)

CONFIG.setdefault("task_modules", [])
# load tasks modules declared as plugin entry points
for entrypoint in pkg_resources.iter_entry_points("swh.workers"):
    worker_registrer_fn = entrypoint.load()
    # The registry function is expected to return a dict which the 'tasks' key
    # is a string (or a list of strings) with the name of the python module in
    # which celery tasks are defined.
    task_modules = worker_registrer_fn().get("task_modules", [])
    CONFIG["task_modules"].extend(task_modules)

# Celery Queues
CELERY_QUEUES = [Queue("celery", Exchange("celery"), routing_key="celery")]

CELERY_DEFAULT_CONFIG = dict(
    # Timezone configuration: all in UTC
    enable_utc=True,
    timezone="UTC",
    # Imported modules
    imports=CONFIG.get("task_modules", []),
    # Time (in seconds, or a timedelta object) for when after stored task
    # tombstones will be deleted. None means to never expire results.
    result_expires=None,
    # A string identifying the default serialization method to use. Can
    # be json (default), pickle, yaml, msgpack, or any custom
    # serialization methods that have been registered with
    task_serializer="json",
    # Result serialization format
    result_serializer="json",
    # Acknowledge tasks as soon as they're received. We can do this as we have
    # external monitoring to decide if we need to retry tasks.
    task_acks_late=False,
    # A string identifying the default serialization method to use.
    # Can be pickle (default), json, yaml, msgpack or any custom serialization
    # methods that have been registered with kombu.serialization.registry
    accept_content=["msgpack", "json"],
    # If True the task will report its status as “started”
    # when the task is executed by a worker.
    task_track_started=True,
    # Default compression used for task messages. Can be gzip, bzip2
    # (if available), or any custom compression schemes registered
    # in the Kombu compression registry.
    # result_compression='bzip2',
    # task_compression='bzip2',
    # Disable all rate limits, even if tasks has explicit rate limits set.
    # (Disabling rate limits altogether is recommended if you don’t have any
    # tasks using them.)
    worker_disable_rate_limits=True,
    # Task routing
    task_routes=route_for_task,
    # Allow pool restarts from remote
    worker_pool_restarts=True,
    # Do not prefetch tasks
    worker_prefetch_multiplier=1,
    # Send events
    worker_send_task_events=True,
    # Do not send useless task_sent events
    task_send_sent_event=False,
)


def build_app(config=None):
    config = merge_configs(
        {k: v for (k, (_, v)) in DEFAULT_CONFIG.items()}, config or {}
    )

    config["task_queues"] = CELERY_QUEUES + [
        Queue(queue, Exchange(queue), routing_key=queue)
        for queue in config.get("task_queues", ())
    ]
    logger.debug("Creating a Celery app with %s", config)

    # Instantiate the Celery app
    app = Celery(broker=config["task_broker"], task_cls="swh.scheduler.task:SWHTask")
    app.add_defaults(CELERY_DEFAULT_CONFIG)
    app.add_defaults(config)
    return app


app = build_app(CONFIG)


@task_prerun.connect
def celery_task_prerun(task_id, task, *args, **kwargs):
    task_sentry_settings = CONFIG.get("sentry_settings_for_celery_tasks", {})
    task_name_parts = task.name.split(".")
    for i in reversed(range(len(task_name_parts) + 1)):
        # sentry settings can be defined for task name or package name
        object_name = ".".join(task_name_parts[:i])
        sentry_settings = task_sentry_settings.get(object_name)
        if sentry_settings:
            _init_sentry(
                sentry_settings.get("dsn"),
                main_package=sentry_settings.get("main_package"),
            )
            break


# XXX for BW compat
Celery.get_queue_length = get_queue_length



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Dict

import pytest
import yaml

from swh.scheduler import DEFAULT_CONFIG
from swh.scheduler.cli.config import read_config


@pytest.fixture
def local_sched_config(swh_scheduler_config):
    """Expose the local scheduler configuration"""
    return {"scheduler": {"cls": "local", **swh_scheduler_config}}


@pytest.fixture
def local_sched_configfile(local_sched_config, tmp_path):
    """Write in temporary location the local scheduler configuration"""
    configfile = tmp_path / "config.yml"
    configfile.write_text(yaml.dump(local_sched_config))
    return configfile.as_posix()


def test_read_config_from_file(local_sched_configfile, local_sched_config):
    """Loading configuration from file should be ok."""

    actual_config = read_config(local_sched_configfile)

    assert actual_config is not None
    assert isinstance(actual_config, Dict)

    assert actual_config == local_sched_config


def test_read_config_from_environment(
    local_sched_configfile, local_sched_config, monkeypatch
):
    """Loading configuration from environment variable should be ok."""

    with monkeypatch.context() as m:
        m.setenv("SWH_CONFIG_FILENAME", local_sched_configfile)
        from os import environ

        assert environ["SWH_CONFIG_FILENAME"] == str(local_sched_configfile)
        actual_config = read_config()

    assert actual_config is not None
    assert isinstance(actual_config, Dict)

    assert actual_config == local_sched_config


def test_read_config_default(local_sched_configfile):
    """Loading configuration from default configuration should be ok."""

    actual_config = read_config()

    assert actual_config is not None
    assert isinstance(actual_config, Dict)

    assert actual_config["scheduler"] == DEFAULT_CONFIG["scheduler"][1]
